#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 25 17:01:24 2019

@author: seykia
@author: augub
"""

from __future__ import print_function
from __future__ import division


import os
import warnings
import sys
import numpy as np
from ast import literal_eval as make_tuple

try:
    from pcntoolkit.dataio import fileio
    from pcntoolkit.normative_model.norm_base import NormBase
    from pcntoolkit.model.hbr import HBR
except ImportError:
    pass

    path = os.path.abspath(os.path.dirname(__file__))
    if path not in sys.path:
        sys.path.append(path)
    del path
    import dataio.fileio as fileio
    from model.hbr import HBR
    from norm_base import NormBase


class NormHBR(NormBase):
        
    """ HBR multi-batch normative modelling class. By default, this function 
    estimates a linear model with random intercept, random slope, and random 
    homoscedastic noise.
    
    :param X: [N×P] array of clinical covariates
    :param y: [N×1] array of neuroimaging measures
    :param trbefile: the address to the batch effects file for the training set.
        the batch effect array should be a [N×M] array where M is the number of
        the type of batch effects. For example when the site and gender is modeled
        as batch effects M=2. Each column in the batch effect array contains the 
        batch ID (starting from 0) for each sample. If not specified (default=None)
        then all samples assumed to be from the same batch (i.e., the batch effect 
                                                            is not modelled).
    :param tsbefile: Similar to trbefile for the test set.
    :param model_type: Specifies the type of the model from 'linear', 'plynomial',
        and 'bspline' (defauls is 'linear').
    :param likelihood: specifies the type of likelihood among 'Normal' 'SHASHb','SHASHo',
        and 'SHASHo2' (defauls is normal).
    :param linear_mu: Boolean (default='True') to decide whether the mean (mu) is
        parametrized on a linear function (thus changes with covariates) or it is fixed.
    :param linear_sigma: Boolean (default='False') to decide whether the variance (sigma) is
        parametrized on a linear function (heteroscedastic noise) or it is fixed for
        each batch (homoscedastic noise).
    :param linear_epsilon: Boolean (default='False') to decide the parametrization
        of epsilon for the SHASH likelihood that controls its skewness. 
        If True, epsilon is  parametrized on a linear function 
        (thus changes with covariates) otherwise it is fixed for each batch.
    :param linear_delta: Boolean (default='False') to decide the parametrization
        of delta for the SHASH likelihood that controls its kurtosis. 
        If True, delta is  parametrized on a linear function 
        (thus changes with covariates) otherwise it is fixed for each batch.
    :param random_intercept_{parameter}: if parameters mu (default='True'), 
        sigma (default='False'), epsilon (default='False'), and delta (default='False')
        are parametrized on a linear function, then this boolean decides 
        whether the intercept can vary across batches.
    :param random_slope_{parameter}: if parameters mu (default='True'), 
        sigma (default='False'), epsilon (default='False'), and delta (default='False')
        are parametrized on a linear function, then this boolean decides 
        whether the slope can vary across batches.
    :param centered_intercept_{parameter}: if parameters mu (default='False'), 
        sigma (default='False'), epsilon (default='False'), and delta (default='False')
        are parametrized on a linear function, then this boolean decides 
        whether the parameters of intercept are estimated in a centered or
        non-centered manner (default). While centered estimation runs faster
        it may cause some problems for the sampler (the funnel of hell).
    :param centered_slope_{parameter}: if parameters mu (default='False'), 
        sigma (default='False'), epsilon (default='False'), and delta (default='False')
        are parametrized on a linear function, then this boolean decides 
        whether the parameters of slope are estimated in a centered or
        non-centered manner (default). While centered estimation runs faster
        it may cause some problems for the sampler (the funnel of hell).
    :param sampler: specifies the type of PyMC sampler (Defauls is 'NUTS').
    :param n_samples: The number of samples to draw (Default is '1000'). Please 
        note that this parameter must be specified in a string fromat ('1000' and 
                                                                       not 1000).  
    :param n_tuning: String that specifies the number of iterations to adjust 
        the samplers's step sizes, scalings or similar (defauls is '500').
    :param n_chains: String that specifies the number of chains to sample. Defauls 
        is '1' for faster estimation, but note that sampling independent chains 
        is important for some convergence checks.
    :param cores: String that specifies the number of chains to run in parallel.
        (defauls is '1').
    :param Initialization method to use for auto-assigned NUTS samplers. The
        defauls is 'jitter+adapt_diag' that starts with a identity mass matrix 
        and then adapt a diagonal based on the variance of the tuning samples
        while adding a uniform jitter in [-1, 1] to the starting point in each chain.
    :param target_accept: String that of a float in [0, 1] that regulates the 
        step size such that we approximate this acceptance rate. The defauls is '0.8'
        but higher values like 0.9 or 0.95 often work better for problematic posteriors.
    :param order: String that defines the order of bspline or polynomial model.
        The defauls is '3'.
    :param nknots: String that defines the numbers of knots for the bspline model.
        The defauls is '5'. Higher values increase the model complexity with negative 
        effect on the spped of estimations.
    :param nn_hidden_layers_num: String the specifies the number of hidden layers
        in neural network model. It can be either '1' or '2'. The default is set to '2'.
    :param nn_hidden_neuron_num: String that specifies the number of neurons in 
        the hidden layers. The defauls is set to  '2'.
        
    Written by S.de Boer and S.M. Kia
    
    """

    def __init__(self, **kwargs):

        self.configs = dict()
        # inputs
        self.configs['trbefile'] = kwargs.pop('trbefile', None)
        self.configs['tsbefile'] = kwargs.pop('tsbefile', None)
        # Model settings
        self.configs['type'] = kwargs.pop('model_type', 'linear')
        self.configs['random_noise'] = kwargs.pop('random_noise', 'True') == 'True'
        self.configs['likelihood'] = kwargs.pop('likelihood', 'Normal')
        # sampler settings
        self.configs['n_samples'] = int(kwargs.pop('n_samples', '1000'))
        self.configs['n_tuning'] = int(kwargs.pop('n_tuning', '500'))
        self.configs['n_chains'] = int(kwargs.pop('n_chains', '1'))
        self.configs['sampler'] = kwargs.pop('sampler', 'NUTS')
        self.configs['target_accept'] = float(kwargs.pop('target_accept', '0.8'))
        self.configs['init'] = kwargs.pop('init', 'jitter+adapt_diag')
        self.configs['cores'] = int(kwargs.pop('cores', '1'))
        # model transfer setting
        self.configs['freedom'] = int(kwargs.pop('freedom', '1'))
        self.configs['transferred'] = False
        # deprecated settings
        self.configs['skewed_likelihood'] = kwargs.pop('skewed_likelihood', 'False') == 'True'
        # misc
        self.configs['pred_type'] = kwargs.pop('pred_type', 'single')

        if self.configs['type'] == 'bspline':
            self.configs['order'] = int(kwargs.pop('order', '3'))
            self.configs['nknots'] = int(kwargs.pop('nknots', '5'))
        elif self.configs['type'] == 'polynomial':
            self.configs['order'] = int(kwargs.pop('order', '3'))
        elif self.configs['type'] == 'nn':
            self.configs['nn_hidden_neuron_num'] = int(kwargs.pop('nn_hidden_neuron_num', '2'))
            self.configs['nn_hidden_layers_num'] = int(kwargs.pop('nn_hidden_layers_num', '2'))
            if self.configs['nn_hidden_layers_num'] > 2:
                raise ValueError("Using " + str(self.configs['nn_hidden_layers_num']) \
                                 + " layers was not implemented. The number of " \
                                 + " layers has to be less than 3.")
        elif self.configs['type'] == 'linear':
            pass
        else:
            raise ValueError("Unknown model type, please specify from 'linear', \
                             'polynomial', 'bspline', or 'nn'.")

        if self.configs['type'] in ['bspline', 'polynomial', 'linear']:

            for p in ['mu', 'sigma', 'epsilon', 'delta']:
                self.configs[f'linear_{p}'] = kwargs.pop(f'linear_{p}', 'False') == 'True'

                ######## Deprecations (remove in later version)
                if f'{p}_linear' in kwargs.keys():
                    print(f'The keyword \'{p}_linear\' is deprecated. It is now automatically replaced with \'linear_{p}\'')
                    self.configs[f'linear_{p}'] = kwargs.pop(f'{p}_linear', 'False') == 'True'
                ##### End Deprecations 

                for c in ['centered','random']:
                    self.configs[f'{c}_{p}'] = kwargs.pop(f'{c}_{p}', 'False') == 'True'
                    for sp in ['slope','intercept']:
                        self.configs[f'{c}_{sp}_{p}'] = kwargs.pop(f'{c}_{sp}_{p}', 'False') == 'True'

            ######## Deprecations (remove in later version)
            if self.configs['linear_sigma']:
                if 'random_noise' in kwargs.keys():
                    print("The keyword \'random_noise\' is deprecated. It is now automatically replaced with \'random_intercept_sigma\', because sigma is linear")
                    self.configs['random_intercept_sigma'] = kwargs.pop('random_noise','True') == 'True'
            elif 'random_noise' in kwargs.keys():
                print("The keyword \'random_noise\' is deprecated. It is now automatically replaced with \'random_sigma\', because sigma is fixed")
                self.configs['random_sigma'] = kwargs.pop('random_noise','True') == 'True'
            if 'random_slope' in kwargs.keys():
                print("The keyword \'random_slope\' is deprecated. It is now automatically replaced with \'random_intercept_mu\'")
                self.configs['random_slope_mu'] = kwargs.pop('random_slope','True') == 'True'
            ##### End Deprecations 


        ## Default parameters
        self.configs['linear_mu'] = kwargs.pop('linear_mu','True') == 'True'
        self.configs['random_mu'] = kwargs.pop('random_mu','True') == 'True'
        self.configs['random_intercept_mu'] = kwargs.pop('random_intercept_mu','True') == 'True'
        self.configs['random_slope_mu'] = kwargs.pop('random_slope_mu','True') == 'True'
        self.configs['random_sigma'] = kwargs.pop('random_sigma','True') == 'True'
        self.configs['centered_sigma'] = kwargs.pop('centered_sigma','True') == 'True'
        ## End default parameters

        self.hbr = HBR(self.configs)

    @property
    def n_params(self):
        return 1

    @property
    def neg_log_lik(self):
        return -1

    def estimate(self, X, y, **kwargs):

        trbefile = kwargs.pop('trbefile', None)
        if trbefile is not None:
            batch_effects_train = fileio.load(trbefile)
        else:
            print('Could not find batch-effects file! Initilizing all as zeros ...')
            batch_effects_train = np.zeros([X.shape[0], 1])

        self.hbr.estimate(X, y, batch_effects_train)

        return self

    def predict(self, Xs, X=None, Y=None, **kwargs):

        tsbefile = kwargs.pop('tsbefile', None)
        if tsbefile is not None:
            batch_effects_test = fileio.load(tsbefile)
        else:
            print('Could not find batch-effects file! Initilizing all as zeros ...')
            batch_effects_test = np.zeros([Xs.shape[0], 1])

        pred_type = self.configs['pred_type']

        if self.configs['transferred'] == False:
            yhat, s2 = self.hbr.predict(Xs, batch_effects_test, pred=pred_type)
        else:
            raise ValueError("This is a transferred model. Please use predict_on_new_sites function.")

        return yhat.squeeze(), s2.squeeze()

    def estimate_on_new_sites(self, X, y, batch_effects):
        self.hbr.estimate_on_new_site(X, y, batch_effects)
        self.configs['transferred'] = True
        return self

    def predict_on_new_sites(self, X, batch_effects):

        yhat, s2 = self.hbr.predict_on_new_site(X, batch_effects)
        return yhat, s2

    def extend(self, X, y, batch_effects, X_dummy_ranges=[[0.1, 0.9, 0.01]],
               merge_batch_dim=0, samples=10, informative_prior=False):

        X_dummy, batch_effects_dummy = self.hbr.create_dummy_inputs(X_dummy_ranges)

        X_dummy, batch_effects_dummy, Y_dummy = self.hbr.generate(X_dummy,
                                                                  batch_effects_dummy, samples)

        batch_effects[:, merge_batch_dim] = batch_effects[:, merge_batch_dim] + \
                                            np.max(batch_effects_dummy[:, merge_batch_dim]) + 1

        if informative_prior:
            self.hbr.adapt(np.concatenate((X_dummy, X)),
                           np.concatenate((Y_dummy, y)),
                           np.concatenate((batch_effects_dummy, batch_effects)))
        else:
            self.hbr.estimate(np.concatenate((X_dummy, X)),
                              np.concatenate((Y_dummy, y)),
                              np.concatenate((batch_effects_dummy, batch_effects)))

        return self

    def tune(self, X, y, batch_effects, X_dummy_ranges=[[0.1, 0.9, 0.01]],
             merge_batch_dim=0, samples=10, informative_prior=False):

        tune_ids = list(np.unique(batch_effects[:, merge_batch_dim]))

        X_dummy, batch_effects_dummy = self.hbr.create_dummy_inputs(X_dummy_ranges)

        for idx in tune_ids:
            X_dummy = X_dummy[batch_effects_dummy[:, merge_batch_dim] != idx, :]
            batch_effects_dummy = batch_effects_dummy[batch_effects_dummy[:, merge_batch_dim] != idx, :]

        X_dummy, batch_effects_dummy, Y_dummy = self.hbr.generate(X_dummy,
                                                                  batch_effects_dummy, samples)

        if informative_prior:
            self.hbr.adapt(np.concatenate((X_dummy, X)),
                           np.concatenate((Y_dummy, y)),
                           np.concatenate((batch_effects_dummy, batch_effects)))
        else:
            self.hbr.estimate(np.concatenate((X_dummy, X)),
                              np.concatenate((Y_dummy, y)),
                              np.concatenate((batch_effects_dummy, batch_effects)))

        return self

    def merge(self, nm, X_dummy_ranges=[[0.1, 0.9, 0.01]], merge_batch_dim=0,
              samples=10):

        X_dummy1, batch_effects_dummy1 = self.hbr.create_dummy_inputs(X_dummy_ranges)
        X_dummy2, batch_effects_dummy2 = nm.hbr.create_dummy_inputs(X_dummy_ranges)

        X_dummy1, batch_effects_dummy1, Y_dummy1 = self.hbr.generate(X_dummy1,
                                                                     batch_effects_dummy1, samples)
        X_dummy2, batch_effects_dummy2, Y_dummy2 = nm.hbr.generate(X_dummy2,
                                                                   batch_effects_dummy2, samples)

        batch_effects_dummy2[:, merge_batch_dim] = batch_effects_dummy2[:, merge_batch_dim] + \
                                                   np.max(batch_effects_dummy1[:, merge_batch_dim]) + 1

        self.hbr.estimate(np.concatenate((X_dummy1, X_dummy2)),
                          np.concatenate((Y_dummy1, Y_dummy2)),
                          np.concatenate((batch_effects_dummy1,
                                          batch_effects_dummy2)))

        return self

    def generate(self, X, batch_effects, samples=10):

        X, batch_effects, generated_samples = self.hbr.generate(X, batch_effects,
                                                                samples)
        return X, batch_effects, generated_samples



from __future__ import print_function

import os
import sys
import numpy as np
import nibabel as nib
import tempfile
import pandas as pd
import re

try:  # run as a package if installed
    from pcntoolkit import configs
except ImportError:
    pass

    path = os.path.abspath(os.path.dirname(__file__))
    path = os.path.dirname(path) # parent directory 
    if path not in sys.path:
        sys.path.append(path)
    del path
    import configs

CIFTI_MAPPINGS = ('dconn', 'dtseries', 'pconn', 'ptseries', 'dscalar',
                  'dlabel', 'pscalar', 'pdconn', 'dpconn',
                  'pconnseries', 'pconnscalar')

CIFTI_VOL_ATLAS = 'Atlas_ROIs.2.nii.gz'

PICKLE_PROTOCOL = configs.PICKLE_PROTOCOL

# ------------------------
# general utility routines
# ------------------------

def predictive_interval(s2_forward,
                        cov_forward,
                        multiplicator):
  # calculates a predictive interval
  
    PI=np.zeros(len(cov_forward))
    for i,xdot in enumerate(cov_forward):
        s=np.sqrt(s2_forward[i])
        PI[i]=multiplicator*s
    return PI

def create_mask(data_array, mask, verbose=False):
    # create a (volumetric) mask either from an input nifti or the nifti itself

    if mask is not None:
        if verbose:
            print('Loading ROI mask ...')
        maskvol = load_nifti(mask, vol=True)
        maskvol = maskvol != 0
    else:
        if len(data_array.shape) < 4:
            dim = data_array.shape[0:3] + (1,)
        else:
            dim = data_array.shape[0:3] + (data_array.shape[3],)

        if verbose:
            print('Generating mask automatically ...')
        if dim[3] == 1:
            maskvol = data_array[:, :, :] != 0
        else:
            maskvol = data_array[:, :, :, 0] != 0

    return maskvol


def vol2vec(dat, mask, verbose=False):
    # vectorise a 3d image

    if len(dat.shape) < 4:
        dim = dat.shape[0:3] + (1,)
    else:
        dim = dat.shape[0:3] + (dat.shape[3],)
        
    #mask = create_mask(dat, mask=mask, verbose=verbose)
    if mask is None:
        mask = create_mask(dat, mask=mask, verbose=verbose)

    # mask the image
    maskid = np.where(mask.ravel())[0]
    dat = np.reshape(dat, (np.prod(dim[0:3]), dim[3]))
    dat = dat[maskid, :]

    # convert to 1-d array if the file only contains one volume
    if dim[3] == 1:
        dat = dat.ravel()

    return dat


def file_type(filename):
    # routine to determine filetype

    if filename.endswith(('.dtseries.nii', '.dscalar.nii', '.dlabel.nii')):
        ftype = 'cifti'
    elif filename.endswith(('.nii.gz', '.nii', '.img', '.hdr')):
        ftype = 'nifti'
    elif filename.endswith(('.txt', '.csv', '.tsv', '.asc')):
        ftype = 'text'
    elif filename.endswith(('.pkl')):
        ftype = 'binary'
    else:
        raise ValueError("I don't know what to do with " + filename)

    return ftype


def file_extension(filename):
    # routine to get the full file extension (e.g. .nii.gz, not just .gz)

    parts = filename.split(os.extsep)

    if parts[-1] == 'gz':
        if parts[-2] == 'nii' or parts[-2] == 'img' or parts[-2] == 'hdr':
            ext = parts[-2] + '.' + parts[-1]
        else:
            ext = parts[-1]
    elif parts[-1] == 'nii':
        if parts[-2] in CIFTI_MAPPINGS:
            ext = parts[-2] + '.' + parts[-1]
        else:
            ext = parts[-1]
    else:
        ext = parts[-1]

    ext = '.' + ext
    return ext


def file_stem(filename):

    idx = filename.find(file_extension(filename))
    stm = filename[0:idx]

    return stm

# --------------
# nifti routines
# --------------


def load_nifti(datafile, mask=None, vol=False, verbose=False):

    if verbose:
        print('Loading nifti: ' + datafile + ' ...')
    img = nib.load(datafile)
    dat = img.get_data()

    if mask is not None:
        mask=load_nifti(mask, vol=True)

    if not vol:
        dat = vol2vec(dat, mask)

    return dat


def save_nifti(data, filename, examplenii, mask, dtype=None):
    
    '''
    Write output to nifti
    
    Basic usage::

        save_nifti(data, filename mask, dtype)

    where the variables are defined below.

    :param data: numpy array containing the data to write out
    :param filename: where to store it
    :param examplenii: nifti to copy the geometry and data type from
    :mask: nifti image containing a mask for the image
    :param dtype: data type for the output image (if different from the image)
    '''
    

    # load mask
    if isinstance(mask, str):
        mask = load_nifti(mask, vol=True)
        mask = mask != 0

    # load example image
    ex_img = nib.load(examplenii)
    ex_img.shape
    dim = ex_img.shape[0:3]
    if len(data.shape) < 2:
        nvol = 1
        data = data[:, np.newaxis]
    else:
        nvol = int(data.shape[1])

    # write data
    array_data = np.zeros((np.prod(dim), nvol))
    array_data[mask.flatten(), :] = data
    array_data = np.reshape(array_data, dim+(nvol,))
    hdr = ex_img.header
    if dtype is not None:
        hdr.set_data_dtype(dtype)
        array_data = array_data.astype(dtype)
    array_img = nib.Nifti1Image(array_data, ex_img.affine, hdr)
    
    nib.save(array_img, filename)

# --------------
# cifti routines
# --------------


def load_cifti(filename, vol=False, mask=None, rmtmp=True):

    # parse the name
    dnam, fnam = os.path.split(filename)
    fpref = file_stem(fnam)
    outstem = os.path.join(tempfile.gettempdir(),
                           str(os.getpid()) + "-" + fpref)

    # extract surface data from the cifti file
    print("Extracting cifti surface data to ", outstem, '-*.func.gii', sep="")
    giinamel = outstem + '-left.func.gii'
    giinamer = outstem + '-right.func.gii'
    os.system('wb_command -cifti-separate ' + filename +
              ' COLUMN -metric CORTEX_LEFT ' + giinamel)
    os.system('wb_command -cifti-separate ' + filename +
              ' COLUMN -metric CORTEX_RIGHT ' + giinamer)

    # load the surface data
    giil = nib.load(giinamel)
    giir = nib.load(giinamer)
    Nimg = len(giil.darrays)
    Nvert = len(giil.darrays[0].data)
    if Nimg == 1:
        out = np.concatenate((giil.darrays[0].data, giir.darrays[0].data),
                             axis=0)
    else:
        Gl = np.zeros((Nvert, Nimg))
        Gr = np.zeros((Nvert, Nimg))
        for i in range(0, Nimg):
            Gl[:, i] = giil.darrays[i].data
            Gr[:, i] = giir.darrays[i].data
        out = np.concatenate((Gl, Gr), axis=0)
    if rmtmp:
        # clean up temporary files
        os.remove(giinamel)
        os.remove(giinamer)

    if vol:
        niiname = outstem + '-vol.nii'
        print("Extracting cifti volume data to ", niiname, sep="")
        os.system('wb_command -cifti-separate ' + filename +
                  ' COLUMN -volume-all ' + niiname)
        vol = load_nifti(niiname, vol=True)
        volmask = create_mask(vol)
        out = np.concatenate((out, vol2vec(vol, volmask)), axis=0)
        if rmtmp:
            os.remove(niiname)

    return out


def save_cifti(data, filename, example, mask=None, vol=True, volatlas=None):
    """ Write output to nifti """

    # do some sanity checks
    if data.dtype == 'float32' or \
       data.dtype == 'float' or \
       data.dtype == 'float64':
        data = data.astype('float32')  # force 32 bit output
        dtype = 'NIFTI_TYPE_FLOAT32'
    else:
        raise(ValueError, 'Only float data types currently handled')

    if len(data.shape) == 1:
        Nimg = 1
        data = data[:, np.newaxis]
    else:
        Nimg = data.shape[1]

    # get the base filename
    dnam, fnam = os.path.split(filename)
    fstem = file_stem(fnam)

    # Split the template
    estem = os.path.join(tempfile.gettempdir(), str(os.getpid()) + "-" + fstem)
    giiexnamel = estem + '-left.func.gii'
    giiexnamer = estem + '-right.func.gii'
    os.system('wb_command -cifti-separate ' + example +
              ' COLUMN -metric CORTEX_LEFT ' + giiexnamel)
    os.system('wb_command -cifti-separate ' + example +
              ' COLUMN -metric CORTEX_RIGHT ' + giiexnamer)

    # write left hemisphere
    giiexl = nib.load(giiexnamel)
    Nvertl = len(giiexl.darrays[0].data)
    garraysl = []
    for i in range(0, Nimg):
        garraysl.append(
            nib.gifti.gifti.GiftiDataArray(data=data[0:Nvertl, i],
            datatype=dtype))
    giil = nib.gifti.gifti.GiftiImage(darrays=garraysl)
    fnamel = fstem + '-left.func.gii'
    nib.save(giil, fnamel)

    # write right hemisphere
    giiexr = nib.load(giiexnamer)
    Nvertr = len(giiexr.darrays[0].data)
    garraysr = []
    for i in range(0, Nimg):
        garraysr.append(
            nib.gifti.gifti.GiftiDataArray(data=data[Nvertl:Nvertl+Nvertr, i],
            datatype=dtype))
    giir = nib.gifti.gifti.GiftiImage(darrays=garraysr)
    fnamer = fstem + '-right.func.gii'
    nib.save(giir, fnamer)

    tmpfiles = [fnamer, fnamel, giiexnamel, giiexnamer]

    # process volumetric data
    if vol:
        niiexname = estem + '-vol.nii'
        os.system('wb_command -cifti-separate ' + example +
                  ' COLUMN -volume-all ' + niiexname)
        niivol = load_nifti(niiexname, vol=True)
        if mask is None:
            mask = create_mask(niivol)

        if volatlas is None:
            volatlas = CIFTI_VOL_ATLAS
        fnamev = fstem + '-vol.nii'

        save_nifti(data[Nvertr+Nvertl:, :], fnamev, niiexname, mask)
        tmpfiles.extend([fnamev, niiexname])

    # write cifti
    fname = fstem + '.dtseries.nii'
    os.system('wb_command -cifti-create-dense-timeseries ' + fname +
              ' -volume ' + fnamev + ' ' + volatlas +
              ' -left-metric ' + fnamel + ' -right-metric ' + fnamer)

    # clean up
    for f in tmpfiles:
        os.remove(f)

# --------------
# ascii routines
# --------------


def load_pd(filename):
    # based on pandas
    x = pd.read_csv(filename,
                    sep=' ',
                    header=None)
    return x


def save_pd(data, filename):
    # based on pandas
    data.to_csv(filename,
                index=None,
                header=None,
                sep=' ',
                na_rep='NaN')


def load_ascii(filename):
    # based on pandas
    x = np.loadtxt(filename)
    return x


def save_ascii(data, filename):
    # based on pandas
    np.savetxt(filename, data)

# ----------------
# generic routines
# ----------------


def save(data, filename, example=None, mask=None, text=False, dtype=None):

    if file_type(filename) == 'cifti':
        save_cifti(data.T, filename, example, vol=True)
    elif file_type(filename) == 'nifti':
        save_nifti(data.T, filename, example, mask, dtype=dtype)
    elif text or file_type(filename) == 'text':
        save_ascii(data, filename)
    elif file_type(filename) == 'binary':
        data = pd.DataFrame(data)
        data.to_pickle(filename, protocol=PICKLE_PROTOCOL)


def load(filename, mask=None, text=False, vol=True):

    if file_type(filename) == 'cifti':
        x = load_cifti(filename, vol=vol)
    elif file_type(filename) == 'nifti':
        x = load_nifti(filename, mask, vol=vol)
    elif text or file_type(filename) == 'text':
        x = load_ascii(filename)
    elif file_type(filename) == 'binary':
        x = pd.read_pickle(filename)
        x = x.to_numpy()

    return x

# -------------------
# sorting routines for batched in normative parallel
# -------------------


def tryint(s):
    try:
        return int(s)
    except ValueError:
        return s


def alphanum_key(s):
    return [tryint(c) for c in re.split('([0-9]+)', s)]


def sort_nicely(l):
    return sorted(l, key=alphanum_key)



from __future__ import print_function
from __future__ import division

import numpy as np
from scipy import optimize , linalg
from scipy.linalg import LinAlgError


class BLR:
    """Bayesian linear regression

    Estimation and prediction of Bayesian linear regression models

    Basic usage::

        B = BLR()
        hyp = B.estimate(hyp0, X, y)
        ys,s2 = B.predict(hyp, X, y, Xs)

    where the variables are

    :param hyp: vector of hyperparmaters.
    :param X: N x D data array
    :param y: 1D Array of targets (length N)
    :param Xs: Nte x D array of test cases
    :param hyp0: starting estimates for hyperparameter optimisation

    :returns: * ys - predictive mean
              * s2 - predictive variance

    The hyperparameters are::

        hyp = ( log(beta), log(alpha) )  # hyp is a list or numpy array

    The implementation and notation mostly follows Bishop (2006).
    The hyperparameter beta is the noise precision and alpha is the precision
    over lengthscale parameters. This can be either a scalar variable (a
    common lengthscale for all input variables), or a vector of length D (a
    different lengthscale for each input variable, derived using an automatic
    relevance determination formulation). These are estimated using conjugate
    gradient optimisation of the marginal likelihood.

    Reference:
    Bishop (2006) Pattern Recognition and Machine Learning, Springer

    Written by A. Marquand
    """

    def __init__(self, **kwargs):
        # parse arguments
        n_iter = kwargs.get('n_iter', 100)
        tol = kwargs.get('tol', 1e-3)
        verbose = kwargs.get('verbose', False)
        var_groups = kwargs.get('var_groups', None)
        var_covariates = kwargs.get('var_covariates', None)
        warp = kwargs.get('warp', None)
        warp_reparam = kwargs.get('warp_reparam', False)
                
        if var_groups is not None and var_covariates is not None:
            raise ValueError("var_covariates and var_groups cannot both be used")
            
        # basic parameters
        self.hyp = np.nan
        self.nlZ = np.nan
        self.tol = tol          # not used at present
        self.n_iter = n_iter
        self.verbose = verbose
        self.var_groups = var_groups
        if var_covariates is not None:
            self.hetero_var = True
        else:
            self.hetero_var = False
        if self.var_groups is not None:
            self.var_ids = set(self.var_groups)
            self.var_ids = sorted(list(self.var_ids))

        # set up warped likelihood
        if verbose:
            print('warp:', warp, 'warp_reparam:', warp_reparam)
        if warp is None:
            self.warp = None
            self.n_warp_param = 0
        else:
            self.warp = warp
            self.n_warp_param = warp.get_n_params()
            self.warp_reparam = warp_reparam
            
        self.gamma = None
    
    def _parse_hyps(self, hyp, X, Xv=None):

        N = X.shape[0]
        
        # noise precision
        if Xv is not None:
            if len(Xv.shape) == 1:
                Dv = 1
                Xv = Xv[:, np.newaxis]
            else:
                Dv = Xv.shape[1]           
            w_d = np.asarray(hyp[0:Dv]) 
            beta = np.exp(Xv.dot(w_d))
            n_lik_param = len(w_d)
        elif self.var_groups is not None:
            beta = np.exp(hyp[0:len(self.var_ids)])
            n_lik_param = len(beta)
        else:
            beta = np.asarray([np.exp(hyp[0])]) 
            n_lik_param = len(beta)
         
        # parameters for warping the likelhood function
        if self.warp is not None:
            gamma = hyp[n_lik_param:(n_lik_param + self.n_warp_param)]
            n_lik_param += self.n_warp_param
        else:
            gamma = None

        # precision for the coefficients
        if isinstance(beta, list) or type(beta) is np.ndarray:
            alpha = np.exp(hyp[n_lik_param:])
        else:
            alpha = np.exp(hyp[1:])

        # reparameterise the warp (WarpSinArcsinh only)
        if self.warp is not None and self.warp_reparam:
            delta = np.exp(gamma[1])
            beta = beta/(delta**2)
    
        # Create precision matrix from noise precision
        if Xv is not None:
            self.lambda_n_vec = beta
        elif self.var_groups is not None:
            beta_all = np.ones(N)
            for v in range(len(self.var_ids)):
                beta_all[self.var_groups == self.var_ids[v]] = beta[v]
            self.lambda_n_vec = beta_all
        else:     
            self.lambda_n_vec = np.ones(N)*beta
    
        return beta, alpha, gamma
        
    def post(self, hyp, X, y, Xv=None):
        """ Generic function to compute posterior distribution.

            This function will save the posterior mean and precision matrix as
            self.m and self.A and will also update internal parameters (e.g.
            N, D and the prior covariance (Sigma_a) and precision (Lambda_a).
            
            :param hyp: hyperparameter vector
            :param X: covariates
            :param y: responses
            :param Xv: covariates for heteroskedastic noise 
        """

        N = X.shape[0]
        if len(X.shape) == 1:
            D = 1
        else:
            D = X.shape[1]

        if (hyp == self.hyp).all() and hasattr(self, 'N'):
            print("hyperparameters have not changed, exiting")
            return
        
        beta, alpha, gamma = self._parse_hyps(hyp, X, Xv)

        if self.verbose:
            print("estimating posterior ... | hyp=", hyp)

        # prior variance
        if len(alpha) == 1 or len(alpha) == D:
            self.Sigma_a = np.diag(np.ones(D))/alpha
            self.Lambda_a = np.diag(np.ones(D))*alpha
        else:
            raise ValueError("hyperparameter vector has invalid length")

        # compute posterior precision and mean
        # this is equivalent to the following operation but makes much more 
        # efficient use of memory by avoiding the need to store Lambda_n
        #
        # self.A = X.T.dot(self.Lambda_n).dot(X) + self.Lambda_a
        # self.m = linalg.solve(self.A, X.T, 
        #                      check_finite=False).dot(self.Lambda_n).dot(y)
        
        XtLambda_n = X.T*self.lambda_n_vec
        self.A = XtLambda_n.dot(X) + self.Lambda_a
        invAXt = linalg.solve(self.A, X.T, check_finite=False)
        self.m = (invAXt*self.lambda_n_vec).dot(y)

        # save stuff
        self.N = N
        self.D = D
        self.hyp = hyp

    def loglik(self, hyp, X, y, Xv=None):
        """ Function to compute compute log (marginal) likelihood """

        # hyperparameters (alpha not needed)
        beta, alpha, gamma = self._parse_hyps(hyp, X, Xv) 

        # warp the likelihood?
        if self.warp is not None:
            if self.verbose:
                print('warping input...')
            y_unwarped = y
            y = self.warp.f(y, gamma)
            
        # load posterior and prior covariance
        if (hyp != self.hyp).any() or not(hasattr(self, 'A')):
            try:
                self.post(hyp, X, y, Xv)
            except ValueError:
                print("Warning: Estimation of posterior distribution failed")
                nlZ = 1/np.finfo(float).eps
                return nlZ
                    
        try:
            # compute the log determinants in a numerically stable way
            logdetA = 2*sum(np.log(np.diag(np.linalg.cholesky(self.A))))
        except (ValueError, LinAlgError):
            print("Warning: Estimation of posterior distribution failed")
            nlZ = 1/np.finfo(float).eps
            return nlZ

        logdetSigma_a = sum(np.log(np.diag(self.Sigma_a))) # diagonal
        logdetSigma_n = sum(np.log(1/self.lambda_n_vec))
        
        # compute negative marginal log likelihood
        X_y_t_sLambda_n = (y-X.dot(self.m))*np.sqrt(self.lambda_n_vec)
        nlZ = -0.5 * (-self.N*np.log(2*np.pi) -
                      logdetSigma_n -
                      logdetSigma_a -
                      X_y_t_sLambda_n.T.dot(X_y_t_sLambda_n) -
                      self.m.T.dot(self.Lambda_a).dot(self.m) -
                      logdetA
                      )
        
        
        if self.warp is not None:
            # add in the Jacobian 
            nlZ = nlZ - sum(np.log(self.warp.df(y_unwarped, gamma)))

        # make sure the output is finite to stop the minimizer getting upset
        if not np.isfinite(nlZ):
            nlZ = 1/np.finfo(float).eps

        if self.verbose:
            print("nlZ= ", nlZ, " | hyp=", hyp)

        self.nlZ = nlZ
        return nlZ
    
    def penalized_loglik(self, hyp, X, y, Xv=None, l=0.1, norm='L1'):
        """ Function to compute the penalized log (marginal) likelihood 
             
            :param hyp: hyperparameter vector
            :param X: covariates
            :param y: responses
            :param Xv: covariates for heteroskedastic noise 
            :param l: regularisation penalty
            :param norm: type of regulariser (L1 or L2)
        """

        if norm.lower() == 'l1':
            L = self.loglik(hyp, X, y, Xv) + l * sum(abs(hyp))
        elif norm.lower() == 'l2':
            L = self.loglik(hyp, X, y, Xv) + l * sum(np.sqrt(hyp**2))
        else:
            print("Requested penalty not recognized, choose between 'L1' or 'L2'.")
        return L

    def dloglik(self, hyp, X, y, Xv=None):
        """ Function to compute derivatives """

        # hyperparameters
        beta, alpha, gamma = self._parse_hyps(hyp, X, Xv)
        
        if self.warp is not None:
            raise ValueError('optimization with derivatives is not yet ' + \
                             'supported for warped liklihood')

        # load posterior and prior covariance
        if (hyp != self.hyp).any() or not(hasattr(self, 'A')):
            try:
                self.post(hyp, X, y, Xv)
            except ValueError:
                print("Warning: Estimation of posterior distribution failed")
                dnlZ = np.sign(self.dnlZ) / np.finfo(float).eps
                return dnlZ

        # precompute re-used quantities to maximise speed 
        # todo: revise implementation to use Cholesky throughout 
        #       that would remove the need to explicitly compute the inverse
        S = np.linalg.inv(self.A)                       # posterior covariance
        SX = S.dot(X.T)
        XLn = X.T*self.lambda_n_vec # = X.T.dot(self.Lambda_n)
        XLny = XLn.dot(y)
        SXLny = S.dot(XLny)       
        XLnXm = XLn.dot(X).dot(self.m)
      
        # initialise derivatives
        dnlZ = np.zeros(hyp.shape)
        dnl2 = np.zeros(hyp.shape)
        
        # noise precision parameter(s)
        for i in range(0, len(beta)):
            # first compute derivative of Lambda_n with respect to beta
            dL_n_vec = np.zeros(self.N)
            if self.var_groups is None:
                dL_n_vec = np.ones(self.N)
            else:    
                dL_n_vec[np.where(self.var_groups == self.var_ids[i])[0]] = 1
            dLambda_n = np.diag(dL_n_vec)
            
            # compute quantities used multiple times
            XdLnX = X.T.dot(dLambda_n).dot(X)
            dA = XdLnX
            
            # derivative of posterior parameters with respect to beta
            b = -S.dot(dA).dot(SXLny) + SX.dot(dLambda_n).dot(y)
            
            # compute np.trace(self.Sigma_n.dot(dLambda_n)) efficiently
            trSigma_ndLambda_n = sum((1/self.lambda_n_vec)*np.diag(dLambda_n))

            # compute  y.T.dot(Lambda_n) efficiently
            ytLn = (y*self.lambda_n_vec).T
            
            # compute derivatives
            dnlZ[i] = - (0.5 * trSigma_ndLambda_n - 
                         0.5 * y.dot(dLambda_n).dot(y) +
                         y.dot(dLambda_n).dot(X).dot(self.m) +
                         ytLn.dot(X).dot(b) -
                         0.5 * self.m.T.dot(XdLnX).dot(self.m) -    
                         b.T.dot(XLnXm) -
                         b.T.dot(self.Lambda_a).dot(self.m) -
                         0.5 * np.trace(S.dot(dA))
                         ) * beta[i]

        # scaling parameter(s)
        for i in range(0, len(alpha)):
            # first compute derivatives with respect to alpha
            if len(alpha) == self.D:  # are we using ARD?
                dLambda_a = np.zeros((self.D, self.D))
                dLambda_a[i, i] = 1
            else:
                dLambda_a = np.eye(self.D)

            F = dLambda_a
            c = -S.dot(F).dot(SXLny)
            
            # compute np.trace(self.Sigma_a.dot(dLambda_a)) efficiently
            trSigma_adLambda_a = sum(np.diag(self.Sigma_a)*np.diag(dLambda_a))
            
            dnlZ[i+len(beta)] = -(0.5* trSigma_adLambda_a +
                                  XLny.T.dot(c) -
                                  c.T.dot(XLnXm) -
                                  c.T.dot(self.Lambda_a).dot(self.m) -
                                  0.5 * self.m.T.dot(F).dot(self.m) -
                                  0.5*np.trace(linalg.solve(self.A, F))
                                  ) * alpha[i]

        # make sure the gradient is finite to stop the minimizer getting upset
        if not all(np.isfinite(dnlZ)):
            bad = np.where(np.logical_not(np.isfinite(dnlZ)))
            for b in bad:
                dnlZ[b] = np.sign(self.dnlZ[b]) / np.finfo(float).eps

        if self.verbose:
            print("dnlZ= ", dnlZ, " | hyp=", hyp)

        self.dnlZ = dnlZ
        return dnlZ

    # model estimation (optimization)
    def estimate(self, hyp0, X, y, **kwargs):
        """ Function to estimate the model 
                 
            :param hyp: hyperparameter vector
            :param X: covariates
            :param y: responses
            :param optimizer: optimisation algorithm ('cg','powell','nelder-mead','l0bfgs-b') 
        """
        
        optimizer = kwargs.get('optimizer','cg')
        
        # covariates for heteroskedastic noise
        Xv = kwargs.get('var_covariates', None)
        
        # options for l-bfgs-b
        l = float(kwargs.get('l', 0.1))
        epsilon = float(kwargs.get('epsilon', 0.1))
        norm = kwargs.get('norm', 'l2')

        if optimizer.lower() == 'cg':  # conjugate gradients
            out = optimize.fmin_cg(self.loglik, hyp0, self.dloglik, (X, y, Xv),
                                   disp=True, gtol=self.tol,
                                   maxiter=self.n_iter, full_output=1)
        elif optimizer.lower() == 'powell':  # Powell's method
            out = optimize.fmin_powell(self.loglik, hyp0, (X, y, Xv),
                                       full_output=1)
        elif optimizer.lower() == 'nelder-mead':
            out = optimize.fmin(self.loglik, hyp0, (X, y, Xv),
                                       full_output=1)
        elif optimizer.lower() == 'l-bfgs-b':
            all_hyp_i = [hyp0]
            def store(X):
                hyp = X
                all_hyp_i.append(hyp)
            try:
                out = optimize.fmin_l_bfgs_b(self.penalized_loglik, x0=hyp0,
                                          args=(X, y, Xv, l, norm), approx_grad=True,
                                          epsilon=epsilon, callback=store)
            # If the matrix becomes singular restart at last found hyp
            except np.linalg.LinAlgError:
                print(f'Restarting estimation at hyp = {all_hyp_i[-1]}, due to *** numpy.linalg.LinAlgError: Matrix is singular.')
                out = optimize.fmin_l_bfgs_b(self.penalized_loglik, x0=all_hyp_i[-1],
                                          args=(X, y, Xv, l, norm), approx_grad=True,
                                          epsilon=epsilon)
        else:
            raise ValueError("unknown optimizer")

        self.hyp = out[0]
        self.nlZ = out[1]
        self.optimizer = optimizer

        return self.hyp

    def predict(self, hyp, X, y, Xs, 
                var_groups_test=None, 
                var_covariates_test=None, **kwargs):
        """ Function to make predictions from the model
             
            :param hyp: hyperparameter vector
            :param X: covariates for training data
            :param y: responses for training data
            :param Xs: covariates for test data
            :param var_covariates_test: test covariates for heteroskedastic noise 
            
            This always returns Gaussian predictions, i.e. 
            
            :returns: * ys - predictive mean
                      * s2 - predictive variance
        """
        
        Xvs = var_covariates_test
        if Xvs is not None and len(Xvs.shape) == 1:
            Xvs = Xvs[:, np.newaxis]
        
        if X is None or y is None:
            # set dummy hyperparameters
            beta, alpha, gamma = self._parse_hyps(hyp, np.zeros((self.N, self.D)), Xvs)
        else:
            
            # set hyperparameters
            beta, alpha, gamma = self._parse_hyps(hyp, X, Xvs)
            
            # do we need to re-estimate the posterior?
            if (hyp != self.hyp).any() or not(hasattr(self, 'A')):
                raise(ValueError, 'posterior not properly estimated')

        N_test = Xs.shape[0]

        ys = Xs.dot(self.m)
        
        if self.var_groups is not None:
            if len(var_groups_test) != N_test:
                raise(ValueError, 'Invalid variance groups for test')
            # separate variance groups
            s2n = np.ones(N_test)
            for v in range(len(self.var_ids)):
                s2n[var_groups_test == self.var_ids[v]] = 1/beta[v]
        else:
            s2n = 1/beta
                    
        # compute xs.dot(S).dot(xs.T) avoiding computing off-diagonal entries
        s2 = s2n + np.sum(Xs*linalg.solve(self.A, Xs.T).T, axis=1)
        
        return ys, s2       
        
    def predict_and_adjust(self, hyp, X, y, Xs=None, 
                             ys=None, 
                             var_groups_test=None,
                             var_groups_adapt=None, **kwargs):
        """ Function to transfer the model to a new site. This is done by
            first making predictions on the adaptation data given by X,
            adjusting by the residuals with respect to y.
            
            :param hyp: hyperparameter vector
            :param X: covariates for adaptation (i.e. calibration) data
            :param y: responses for adaptation data
            :param Xs: covariate data (for which predictions should be adjusted)
            :param ys: true response variables (to be adjusted)
            :param var_groups_test: variance groups (e.g. sites) for test data
            :param var_groups_adapt: variance groups for adaptation data
            
            There are two possible ways of using this function, depending on
            whether ys or Xs is specified
                
            If ys is specified, this is applied directly to the data, which is
            assumed to be in the input space (i.e. not warped). In this case 
            the adjusted true data points are returned in the same space
            
            Alternatively, Xs is specified, then the predictions are made and 
            adjusted. In this case the predictive variance are returned in the 
            warped (i.e. Gaussian) space.
            
            This function needs to know which sites are associated with which 
            data points, which provided by var_groups_xxx, which is a list or 
            array of scalar ids .
        """
        
        if ys is None:
            if Xs is None:
                raise ValueError('Either ys or Xs must be specified')
            else:
                N = Xs.shape[0]
        else:
            if len(ys.shape) < 1:
                raise ValueError('ys is specified but has insufficent length')
            N = ys.shape[0]
        
        if var_groups_test is None: 
            var_groups_test = np.ones(N)
            var_groups_adapt = np.ones(X.shape[0])
        
        ys_out = np.zeros(N)
        s2_out = np.zeros(N)
        for g in np.unique(var_groups_test):
            idx_s = var_groups_test == g
            idx_a = var_groups_adapt == g
            
            if sum(idx_a) < 2:
                raise ValueError('Insufficient adaptation data to estimate variance')
            
            # Get predictions from old model on new data X
            ys_ref, s2_ref = self.predict(hyp, None, None, X[idx_a,:])
        
            # Subtract the predictions from true data to get the residuals
            if self.warp is None:
                residuals = ys_ref-y[idx_a]
            else:
                 # Calculate the residuals in warped space
                 y_ref_ws = self.warp.f(y[idx_a], hyp[1:self.warp.get_n_params()+1])
                 residuals = ys_ref - y_ref_ws 
            
            residuals_mu = np.mean(residuals)
            residuals_sd = np.std(residuals)

            # Adjust the mean with the mean of the residuals
            if ys is None:
                # make and adjust predictions
                ys_out[idx_s], s2_out[idx_s] = self.predict(hyp, None, None, Xs[idx_s,:])
                ys_out[idx_s] = ys_out[idx_s] - residuals_mu 
                
                # Set the deviation to the devations of the residuals
                s2_out[idx_s] = np.ones(len(s2_out[idx_s]))*residuals_sd**2
            else:
                # adjust the data 
                if self.warp is not None:
                    y_ws = self.warp.f(ys[idx_s], hyp[1:self.warp.get_n_params()+1])
                    ys_out[idx_s] = y_ws + residuals_mu 
                    ys_out[idx_s] = self.warp.invf(ys_out[idx_s], hyp[1:self.warp.get_n_params()+1])
                else:
                    ys = ys - residuals_mu    
                s2_out = None

        return ys_out, s2_out
        



#!/opt/conda/bin/python

# ------------------------------------------------------------------------------
#  Usage:
#  python normative.py -m [maskfile] -k [number of CV folds] -c <covariates>
#                      -t [test covariates] -r [test responses] <infile>
#
#  Either the -k switch or -t switch should be specified, but not both.
#  If -t is selected, a set of responses should be provided with the -r switch
#
#  Written by A. Marquand
# ------------------------------------------------------------------------------

from __future__ import print_function
from __future__ import division

import os
import sys
import numpy as np
import argparse
import pickle
import glob

from sklearn.model_selection import KFold
from pathlib import Path
            
try:  # run as a package if installed
    from pcntoolkit import configs
    from pcntoolkit.dataio import fileio
    from pcntoolkit.normative_model.norm_utils import norm_init
    from pcntoolkit.util.utils import compute_pearsonr, CustomCV, explained_var
    from pcntoolkit.util.utils import compute_MSLL, scaler, get_package_versions
except ImportError:
    pass

    path = os.path.abspath(os.path.dirname(__file__))
    if path not in sys.path:
        sys.path.append(path)
        #sys.path.append(os.path.join(path,'normative_model'))
    del path
    
    import configs
    from dataio import fileio

    from util.utils import compute_pearsonr, CustomCV, explained_var, compute_MSLL
    from util.utils import scaler, get_package_versions
    from normative_model.norm_utils import norm_init

PICKLE_PROTOCOL = configs.PICKLE_PROTOCOL

def load_response_vars(datafile, maskfile=None, vol=True):
    """ load response variables (of any data type)"""

    if fileio.file_type(datafile) == 'nifti':
        dat = fileio.load_nifti(datafile, vol=vol)
        volmask = fileio.create_mask(dat, mask=maskfile)
        Y = fileio.vol2vec(dat, volmask).T
    else:
        Y = fileio.load(datafile)
        volmask = None
        if fileio.file_type(datafile) == 'cifti':
            Y = Y.T

    return Y, volmask


def get_args(*args):
    """ Parse command line arguments"""

    # parse arguments
    parser = argparse.ArgumentParser(description="Normative Modeling")
    parser.add_argument("responses")
    parser.add_argument("-f", help="Function to call", dest="func", 
                        default="estimate")
    parser.add_argument("-m", help="mask file", dest="maskfile", default=None)
    parser.add_argument("-c", help="covariates file", dest="covfile",
                        default=None)
    parser.add_argument("-k", help="cross-validation folds", dest="cvfolds",
                        default=None)
    parser.add_argument("-t", help="covariates (test data)", dest="testcov",
                        default=None)
    parser.add_argument("-r", help="responses (test data)", dest="testresp",
                        default=None)
    parser.add_argument("-a", help="algorithm", dest="alg", default="gpr")
    parser.add_argument("-x", help="algorithm specific config options", 
                        dest="configparam", default=None)
    # parser.add_argument('-s', action='store_false', 
    #                 help="Flag to skip standardization.", dest="standardize")
    parser.add_argument("keyword_args", nargs=argparse.REMAINDER)
    
    args = parser.parse_args()
    
    # Process required  arguemnts 
    wdir = os.path.realpath(os.path.curdir)
    respfile = os.path.join(wdir, args.responses)
    if args.covfile is None:
        raise(ValueError, "No covariates specified")
    else:
        covfile = args.covfile
    
    # Process optional arguments
    if args.maskfile is None:
        maskfile = None
    else:
        maskfile = os.path.join(wdir, args.maskfile)
    if args.testcov is None and args.cvfolds is not None:
        testcov = None
        testresp = None
        cvfolds = int(args.cvfolds)
        print("Running under " + str(cvfolds) + " fold cross-validation.")
    else:
        print("Test covariates specified")
        testcov = args.testcov
        cvfolds = None
        if args.testresp is None:
            testresp = None
            print("No test response variables specified")
        else:
            testresp = args.testresp
        if args.cvfolds is not None:
            print("Ignoring cross-valdation specification (test data given)")

    # Process addtional keyword arguments. These are always added as strings
    kw_args = {}
    for kw in args.keyword_args:
        kw_arg = kw.split('=')
    
        exec("kw_args.update({'" +  kw_arg[0] + "' : " + 
                              "'" + str(kw_arg[1]) + "'" + "})")
    
    return respfile, maskfile, covfile, cvfolds, \
            testcov, testresp, args.func, args.alg, \
            args.configparam, kw_args
            

def evaluate(Y, Yhat, S2=None, mY=None, sY=None, nlZ=None, nm=None, Xz_tr=None, alg=None,
             metrics = ['Rho', 'RMSE', 'SMSE', 'EXPV', 'MSLL']):
    ''' Compute error metrics
    This function will compute error metrics based on a set of predictions Yhat
    and a set of true response variables Y, namely:
    
    * Rho: Pearson correlation
    * RMSE: root mean squared error
    * SMSE: standardized mean squared error
    * EXPV: explained variance
        
    If the predictive variance is also specified the log loss will be computed
    (which also takes into account the predictive variance). If the mean and 
    standard deviation are also specified these will be used to standardize 
    this, yielding the mean standardized log loss
    
    :param Y: N x P array of true response variables
    :param Yhat: N x P array of predicted response variables
    :param S2: predictive variance
    :param mY: mean of the training set
    :param sY: standard deviation of the training set

    :returns metrics: evaluation metrics
    
    '''
    
    feature_num = Y.shape[1]
    
    # Remove metrics that cannot be computed with only a single data point 
    if Y.shape[0] == 1:
        if 'MSLL' in metrics:
            metrics.remove('MSLL')
        if 'SMSE' in metrics:
            metrics.remove('SMSE')
    
    # find and remove bad variables from the response variables
    nz = np.where(np.bitwise_and(np.isfinite(Y).any(axis=0),
                                 np.var(Y, axis=0) != 0))[0]
    
    MSE = np.mean((Y - Yhat)**2, axis=0)
    
    results = dict()
    
    if 'RMSE' in metrics:
        RMSE = np.sqrt(MSE)
        results['RMSE'] = RMSE
    
    if 'Rho' in metrics:
        Rho = np.zeros(feature_num)
        pRho = np.ones(feature_num)    
        Rho[nz], pRho[nz] = compute_pearsonr(Y[:,nz], Yhat[:,nz])
        results['Rho'] = Rho
        results['pRho'] = pRho
        
    if 'SMSE' in metrics:
        SMSE = np.zeros_like(MSE)
        SMSE[nz] = MSE[nz] / np.var(Y[:,nz], axis=0)
        results['SMSE'] = SMSE
    
    if 'EXPV' in metrics:
        EXPV = np.zeros(feature_num)
        EXPV[nz] = explained_var(Y[:,nz], Yhat[:,nz])
        results['EXPV'] = EXPV
        
    if 'MSLL' in metrics:
        if ((S2 is not None) and (mY is not None) and (sY is not None)):
            MSLL = np.zeros(feature_num)
            MSLL[nz] = compute_MSLL(Y[:,nz], Yhat[:,nz], S2[:,nz], 
                                    mY.reshape(-1,1).T, 
                                    (sY**2).reshape(-1,1).T)
            results['MSLL'] = MSLL
            
    if 'NLL' in metrics:
        results['NLL'] = nlZ
    
    if 'BIC' in metrics:
        if hasattr(getattr(nm, alg), 'hyp'):
            n = Xz_tr.shape[0]
            k = len(getattr(nm, alg).hyp)
            BIC = k * np.log(n) + 2 * nlZ
            results['BIC'] = BIC    
    
    return results

def save_results(respfile, Yhat, S2, maskvol, Z=None, outputsuffix=None, 
                 results=None, save_path=''):
    
    print("Writing outputs ...")
    if respfile is None:
        exfile = None
        file_ext = '.pkl'
    else:
        if fileio.file_type(respfile) == 'cifti' or \
           fileio.file_type(respfile) == 'nifti':
            exfile = respfile
        else:
            exfile = None
        file_ext = fileio.file_extension(respfile)

    if outputsuffix is not None:
        ext = str(outputsuffix) + file_ext
    else:
        ext = file_ext

    fileio.save(Yhat, os.path.join(save_path, 'yhat' + ext), example=exfile, 
                                   mask=maskvol)
    fileio.save(S2, os.path.join(save_path, 'ys2' + ext), example=exfile, 
                mask=maskvol)
    if Z is not None:
        fileio.save(Z, os.path.join(save_path, 'Z' + ext), example=exfile, 
                    mask=maskvol)

    if results is not None:        
        for metric in list(results.keys()):
            if (metric == 'NLL' or metric == 'BIC') and file_ext == '.nii.gz':
                fileio.save(results[metric], os.path.join(save_path, metric + str(outputsuffix) + '.pkl'), 
                        example=exfile, mask=maskvol)
            else:
                fileio.save(results[metric], os.path.join(save_path, metric + ext), 
                            example=exfile, mask=maskvol)

def estimate(covfile, respfile, **kwargs):
    """ Estimate a normative model

    This will estimate a model in one of two settings according to 
    theparticular parameters specified (see below)
        
    * under k-fold cross-validation.
      requires respfile, covfile and cvfolds>=2
    * estimating a training dataset then applying to a second test dataset.
      requires respfile, covfile, testcov and testresp.
    * estimating on a training dataset ouput of forward maps mean and se. 
      requires respfile, covfile and testcov

    The models are estimated on the basis of data stored on disk in ascii or
    neuroimaging data formats (nifti or cifti). Ascii data should be in
    tab or space delimited format with the number of subjects in rows and the
    number of variables in columns. Neuroimaging data will be reshaped
    into the appropriate format

    Basic usage::

        estimate(covfile, respfile, [extra_arguments])

    where the variables are defined below. Note that either the cfolds
    parameter or (testcov, testresp) should be specified, but not both.

    :param respfile: response variables for the normative model
    :param covfile: covariates used to predict the response variable
    :param maskfile: mask used to apply to the data (nifti only)
    :param cvfolds: Number of cross-validation folds
    :param testcov: Test covariates
    :param testresp: Test responses
    :param alg: Algorithm for normative model
    :param configparam: Parameters controlling the estimation algorithm
    :param saveoutput: Save the output to disk? Otherwise returned as arrays
    :param outputsuffix: Text string to add to the output filenames
    :param inscale: Scaling approach for input covariates, could be 'None' (Default), 
                    'standardize', 'minmax', or 'robminmax'.
    :param outscale: Scaling approach for output responses, could be 'None' (Default), 
                    'standardize', 'minmax', or 'robminmax'.

    All outputs are written to disk in the same format as the input. These are:

    :outputs: * yhat - predictive mean
              * ys2 - predictive variance
              * nm - normative model
              * Z - deviance scores
              * Rho - Pearson correlation between true and predicted responses
              * pRho - parametric p-value for this correlation
              * rmse - root mean squared error between true/predicted responses
              * smse - standardised mean squared error

    The outputsuffix may be useful to estimate multiple normative models in the
    same directory (e.g. for custom cross-validation schemes)
    """
    
    # parse keyword arguments 
    maskfile = kwargs.pop('maskfile',None)
    cvfolds = kwargs.pop('cvfolds', None)
    testcov = kwargs.pop('testcov', None)
    testresp = kwargs.pop('testresp',None)
    alg = kwargs.pop('alg','gpr')
    outputsuffix = kwargs.pop('outputsuffix','estimate')
    outputsuffix = "_" + outputsuffix.replace("_", "")  # Making sure there is only one 
                                                        # '_' is in the outputsuffix to 
                                                        # avoid file name parsing problem.
    inscaler = kwargs.pop('inscaler','None')
    outscaler = kwargs.pop('outscaler','None')
    warp = kwargs.get('warp', None)

    # convert from strings if necessary
    saveoutput = kwargs.pop('saveoutput','True')
    if type(saveoutput) is str:
        saveoutput = saveoutput=='True'
    savemodel = kwargs.pop('savemodel','False')
    if type(savemodel) is str:
        savemodel = savemodel=='True'
    
    if savemodel and not os.path.isdir('Models'):
        os.mkdir('Models')

    # which output metrics to compute
    metrics = ['Rho', 'RMSE', 'SMSE', 'EXPV', 'MSLL','NLL', 'BIC']
    
    # load data
    print("Processing data in " + respfile)
    X = fileio.load(covfile)
    Y, maskvol = load_response_vars(respfile, maskfile)
    if len(Y.shape) == 1:
        Y = Y[:, np.newaxis]
    if len(X.shape) == 1:
        X = X[:, np.newaxis]
    Nmod = Y.shape[1]
    
    if (testcov is not None) and (cvfolds is None): # a separate test dataset
        
        run_cv = False
        cvfolds = 1
        Xte = fileio.load(testcov)
        if len(Xte.shape) == 1:
            Xte = Xte[:, np.newaxis]
        if testresp is not None:
            Yte, testmask = load_response_vars(testresp, maskfile)
            if len(Yte.shape) == 1:
                Yte = Yte[:, np.newaxis]
        else:
            sub_te = Xte.shape[0]
            Yte = np.zeros([sub_te, Nmod])
            
        # treat as a single train-test split
        testids = range(X.shape[0], X.shape[0]+Xte.shape[0])
        splits = CustomCV((range(0, X.shape[0]),), (testids,))

        Y = np.concatenate((Y, Yte), axis=0)
        X = np.concatenate((X, Xte), axis=0)
        
    else:
        run_cv = True
        # we are running under cross-validation
        splits = KFold(n_splits=cvfolds, shuffle=True)
        testids = range(0, X.shape[0])
        if alg=='hbr':
           trbefile = kwargs.get('trbefile', None) 
           if trbefile is not None:
                be = fileio.load(trbefile)
                if len(be.shape) == 1:
                    be = be[:, np.newaxis]
           else:
                print('No batch-effects file! Initilizing all as zeros!')
                be = np.zeros([X.shape[0],1])

    # find and remove bad variables from the response variables
    # note: the covariates are assumed to have already been checked
    nz = np.where(np.bitwise_and(np.isfinite(Y).any(axis=0),
                                 np.var(Y, axis=0) != 0))[0]

    # run cross-validation loop
    Yhat = np.zeros_like(Y)
    S2 = np.zeros_like(Y)
    Z = np.zeros_like(Y)
    nlZ = np.zeros((Nmod, cvfolds))
    
    scaler_resp = []
    scaler_cov = []
    mean_resp = [] # this is just for computing MSLL
    std_resp = [] # this is just for computing MSLL
    
    if warp is not None:
        Ywarp = np.zeros_like(Yhat)
        
        # for warping we need to compute metrics separately for each fold
        results_folds = dict()
        for m in metrics:
            results_folds[m]= np.zeros((Nmod, cvfolds))

    for idx in enumerate(splits.split(X)):

        fold = idx[0]
        tr = idx[1][0]
        ts = idx[1][1]

        # standardize responses and covariates, ignoring invalid entries
        iy_tr, jy_tr = np.ix_(tr, nz)
        iy_ts, jy_ts = np.ix_(ts, nz)
        mY = np.mean(Y[iy_tr, jy_tr], axis=0)
        sY = np.std(Y[iy_tr, jy_tr], axis=0)
        mean_resp.append(mY)
        std_resp.append(sY)
        
        if inscaler in ['standardize', 'minmax', 'robminmax']:
            X_scaler = scaler(inscaler)
            Xz_tr = X_scaler.fit_transform(X[tr, :])
            Xz_ts = X_scaler.transform(X[ts, :])
            scaler_cov.append(X_scaler)
        else:
            Xz_tr = X[tr, :]
            Xz_ts = X[ts, :]
            
        if outscaler in ['standardize', 'minmax', 'robminmax']:
            Y_scaler = scaler(outscaler)
            Yz_tr = Y_scaler.fit_transform(Y[iy_tr, jy_tr])
            scaler_resp.append(Y_scaler)
        else:
            Yz_tr = Y[iy_tr, jy_tr]
        
        if (run_cv==True and alg=='hbr'):
            fileio.save(be[tr,:], 'be_kfold_tr_tempfile.pkl')
            fileio.save(be[ts,:], 'be_kfold_ts_tempfile.pkl')
            kwargs['trbefile'] = 'be_kfold_tr_tempfile.pkl'
            kwargs['tsbefile'] = 'be_kfold_ts_tempfile.pkl'

        # estimate the models for all subjects
        for i in range(0, len(nz)):              
            print("Estimating model ", i+1, "of", len(nz))
            nm = norm_init(Xz_tr, Yz_tr[:, i], alg=alg, **kwargs)
                
            try:
                nm = nm.estimate(Xz_tr, Yz_tr[:, i], **kwargs)     
                yhat, s2 = nm.predict(Xz_ts, Xz_tr, Yz_tr[:, i], **kwargs)
                
                if savemodel:
                    nm.save('Models/NM_' + str(fold) + '_' + str(nz[i]) + 
                            outputsuffix + '.pkl' )
                
                if outscaler == 'standardize': 
                    Yhat[ts, nz[i]] = Y_scaler.inverse_transform(yhat, index=i)
                    S2[ts, nz[i]] = s2 * sY[i]**2
                elif outscaler in ['minmax', 'robminmax']:
                    Yhat[ts, nz[i]] = Y_scaler.inverse_transform(yhat, index=i)
                    S2[ts, nz[i]] = s2 * (Y_scaler.max[i] - Y_scaler.min[i])**2
                else:
                    Yhat[ts, nz[i]] = yhat
                    S2[ts, nz[i]] = s2
                    
                nlZ[nz[i], fold] = nm.neg_log_lik
                
                if (run_cv or testresp is not None):
                    if warp is not None:
                        # TODO: Warping for scaled data
                        if outscaler is not None and outscaler != 'None':
                            raise(ValueError, "outscaler not yet supported warping")
                        warp_param = nm.blr.hyp[1:nm.blr.warp.get_n_params()+1] 
                        Ywarp[ts, nz[i]] = nm.blr.warp.f(Y[ts, nz[i]], warp_param)
                        Ytest = Ywarp[ts, nz[i]]
                        
                        # Save warped mean of the training data (for MSLL)
                        yw = nm.blr.warp.f(Y[tr, nz[i]], warp_param)
                        
                        # create arrays for evaluation
                        Yhati = Yhat[ts, nz[i]]
                        Yhati = Yhati[:, np.newaxis]
                        S2i = S2[ts, nz[i]]
                        S2i = S2i[:, np.newaxis]

                        # evaluate and save results
                        mf = evaluate(Ytest[:, np.newaxis], Yhati, S2=S2i, 
                                      mY=np.mean(yw), sY=np.std(yw), 
                                      nlZ=nm.neg_log_lik, nm=nm, Xz_tr=Xz_tr, 
                                      alg=alg, metrics = metrics)
                        for k in metrics:
                            results_folds[k][nz[i]][fold] = mf[k]
                    else:
                        Ytest = Y[ts, nz[i]] 
                    
                    Z[ts, nz[i]] = (Ytest - Yhat[ts, nz[i]]) / \
                                    np.sqrt(S2[ts, nz[i]])       
                    
            except Exception as e:
                exc_type, exc_obj, exc_tb = sys.exc_info()
                fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
                print("Model ", i+1, "of", len(nz),
                      "FAILED!..skipping and writing NaN to outputs")
                print("Exception:")
                print(e)
                print(exc_type, fname, exc_tb.tb_lineno)

                Yhat[ts, nz[i]] = float('nan')
                S2[ts, nz[i]] = float('nan')
                nlZ[nz[i], fold] = float('nan')
                if testcov is None:
                    Z[ts, nz[i]] = float('nan')
                else:
                    if testresp is not None:
                        Z[ts, nz[i]] = float('nan')


    if savemodel:
        print('Saving model meta-data...')
        v  = get_package_versions()
        with open('Models/meta_data.md', 'wb') as file:
            pickle.dump({'valid_voxels':nz, 'fold_num':cvfolds, 
                         'mean_resp':mean_resp, 'std_resp':std_resp, 
                         'scaler_cov':scaler_cov, 'scaler_resp':scaler_resp, 
                         'regressor':alg, 'inscaler':inscaler, 
                         'outscaler':outscaler, 'versions':v}, 
                        file, protocol=PICKLE_PROTOCOL)    

    # compute performance metrics
    if (run_cv or testresp is not None):
        print("Evaluating the model ...")
        if warp is None:
            results = evaluate(Y[testids, :], Yhat[testids, :], 
                               S2=S2[testids, :], mY=mean_resp[0], 
                               sY=std_resp[0], nlZ=nlZ, nm=nm, Xz_tr=Xz_tr, alg=alg,
                               metrics = metrics)
        else:
            # for warped data we just aggregate across folds
            results = dict()
            for m in ['Rho', 'RMSE', 'SMSE', 'EXPV', 'MSLL']:
                results[m] = np.mean(results_folds[m], axis=1)
            results['NLL'] = results_folds['NLL']
            results['BIC'] = results_folds['BIC']            
        
    # Set writing options
    if saveoutput:
        if (run_cv or testresp is not None):
            save_results(respfile, Yhat[testids, :], S2[testids, :], maskvol, 
                         Z=Z[testids, :], results=results, 
                         outputsuffix=outputsuffix)
            
        else:
            save_results(respfile, Yhat[testids, :], S2[testids, :], maskvol,
                         outputsuffix=outputsuffix)
                
    else:
        if (run_cv or testresp is not None):
            output = (Yhat[testids, :], S2[testids, :], nm, Z[testids, :], 
                      results)
        else:
            output = (Yhat[testids, :], S2[testids, :], nm)
        
        return output


def fit(covfile, respfile, **kwargs):
    
    # parse keyword arguments 
    maskfile = kwargs.pop('maskfile',None)
    alg = kwargs.pop('alg','gpr')
    savemodel = kwargs.pop('savemodel','True')=='True'
    outputsuffix = kwargs.pop('outputsuffix','fit')
    outputsuffix = "_" + outputsuffix.replace("_", "")
    inscaler = kwargs.pop('inscaler','None')
    outscaler = kwargs.pop('outscaler','None')
    
    if savemodel and not os.path.isdir('Models'):
        os.mkdir('Models')

    # load data
    print("Processing data in " + respfile)
    X = fileio.load(covfile)
    Y, maskvol = load_response_vars(respfile, maskfile)
    if len(Y.shape) == 1:
        Y = Y[:, np.newaxis]
    if len(X.shape) == 1:
        X = X[:, np.newaxis]
    
    # find and remove bad variables from the response variables
    # note: the covariates are assumed to have already been checked
    nz = np.where(np.bitwise_and(np.isfinite(Y).any(axis=0),
                                 np.var(Y, axis=0) != 0))[0]        
    
    scaler_resp = []
    scaler_cov = []
    mean_resp = [] # this is just for computing MSLL
    std_resp = []   # this is just for computing MSLL
    
    # standardize responses and covariates, ignoring invalid entries
    mY = np.mean(Y[:, nz], axis=0)
    sY = np.std(Y[:, nz], axis=0)
    mean_resp.append(mY)
    std_resp.append(sY)
    
    if inscaler in ['standardize', 'minmax', 'robminmax']:
        X_scaler = scaler(inscaler)
        Xz = X_scaler.fit_transform(X)
        scaler_cov.append(X_scaler)
    else:
        Xz = X
        
    if outscaler in ['standardize', 'minmax', 'robminmax']:
        Yz = np.zeros_like(Y)
        Y_scaler = scaler(outscaler)
        Yz[:, nz] = Y_scaler.fit_transform(Y[:, nz])
        scaler_resp.append(Y_scaler)
    else:
        Yz = Y

    # estimate the models for all subjects
    for i in range(0, len(nz)):  
        print("Estimating model ", i+1, "of", len(nz))
        nm = norm_init(Xz, Yz[:, nz[i]], alg=alg, **kwargs)
        nm = nm.estimate(Xz, Yz[:, nz[i]], **kwargs)     
            
        if savemodel:
            nm.save('Models/NM_' + str(0) + '_' + str(nz[i]) + outputsuffix + 
                    '.pkl' )

    if savemodel:
        print('Saving model meta-data...')
        v  = get_package_versions()
        with open('Models/meta_data.md', 'wb') as file:
            pickle.dump({'valid_voxels':nz,
                         'mean_resp':mean_resp, 'std_resp':std_resp, 
                         'scaler_cov':scaler_cov, 'scaler_resp':scaler_resp, 
                         'regressor':alg, 'inscaler':inscaler,
                         'outscaler':outscaler, 'versions':v}, 
                        file, protocol=PICKLE_PROTOCOL)
        
    return nm

    
def predict(covfile, respfile, maskfile=None, **kwargs):
    '''
    Make predictions on the basis of a pre-estimated normative model 
    If only the covariates are specified then only predicted mean and variance 
    will be returned. If the test responses are also specified then quantities
    That depend on those will also be returned (Z scores and error metrics)

    Basic usage::

        predict(covfile, [extra_arguments])

    where the variables are defined below.

    :param covfile: test covariates used to predict the response variable
    :param respfile: test response variables for the normative model
    :param maskfile: mask used to apply to the data (nifti only)
    :param model_path: Directory containing the normative model and metadata.
     When using parallel prediction, do not pass the model path. It will be 
     automatically decided.
    :param outputsuffix: Text string to add to the output filenames
    :param batch_size: batch size (for use with normative_parallel)
    :param job_id: batch id
    :param fold: which cross-validation fold to use (default = 0)
    :param fold: list of model IDs to predict (if not specified all are computed)

    All outputs are written to disk in the same format as the input. These are:

    :outputs: * Yhat - predictive mean
              * S2 - predictive variance
              * Z - Z scores
    '''
    
    
    model_path = kwargs.pop('model_path', 'Models')
    job_id = kwargs.pop('job_id', None)
    batch_size = kwargs.pop('batch_size', None)
    outputsuffix = kwargs.pop('outputsuffix', 'predict')
    outputsuffix = "_" + outputsuffix.replace("_", "")
    inputsuffix = kwargs.pop('inputsuffix', 'estimate')
    inputsuffix = "_" + inputsuffix.replace("_", "")
    alg = kwargs.pop('alg')
    fold = kwargs.pop('fold',0)
    models = kwargs.pop('models', None)
    
    if alg == 'gpr':
        raise(ValueError, "gpr is not supported with predict()")
        
    if respfile is not None and not os.path.exists(respfile):
        print("Response file does not exist. Only returning predictions")
        respfile = None
    if not os.path.isdir(model_path):
        print('Models directory does not exist!')
        return
    else:
        if os.path.exists(os.path.join(model_path, 'meta_data.md')):
            with open(os.path.join(model_path, 'meta_data.md'), 'rb') as file:
                meta_data = pickle.load(file)
            inscaler = meta_data['inscaler']
            outscaler = meta_data['outscaler']
            mY = meta_data['mean_resp']
            sY = meta_data['std_resp']
            scaler_cov = meta_data['scaler_cov']
            scaler_resp = meta_data['scaler_resp']
            meta_data = True
        else:
            print("No meta-data file is found!")
            inscaler = 'None'
            outscaler = 'None'
            meta_data = False

    if batch_size is not None:
        batch_size = int(batch_size)
        job_id = int(job_id) - 1

    
    # load data
    print("Loading data ...")
    X = fileio.load(covfile)
    if len(X.shape) == 1:
        X = X[:, np.newaxis]
    
    sample_num = X.shape[0]
    if models is not None:
        feature_num = len(models)
    else:
        feature_num = len(glob.glob(os.path.join(model_path, 'NM_'+ str(fold) +
                                                 '_*' + inputsuffix + '.pkl')))
        models = range(feature_num)

    Yhat = np.zeros([sample_num, feature_num])
    S2 = np.zeros([sample_num, feature_num])
    Z = np.zeros([sample_num, feature_num])
    
    if inscaler in ['standardize', 'minmax', 'robminmax']:
        Xz = scaler_cov[fold].transform(X)
    else:
        Xz = X
    
    # estimate the models for all subjects
    for i, m in enumerate(models):
        print("Prediction by model ", i+1, "of", feature_num)      
        nm = norm_init(Xz)
        nm = nm.load(os.path.join(model_path, 'NM_' + str(fold) + '_' + 
                                  str(m) + inputsuffix + '.pkl'))
        if (alg!='hbr' or nm.configs['transferred']==False):
            yhat, s2 = nm.predict(Xz, **kwargs)
        else:
            tsbefile = kwargs.get('tsbefile') 
            batch_effects_test = fileio.load(tsbefile)
            yhat, s2 = nm.predict_on_new_sites(Xz, batch_effects_test)
        
        if outscaler == 'standardize': 
            Yhat[:, i] = scaler_resp[fold].inverse_transform(yhat, index=i)
            S2[:, i] = s2.squeeze() * sY[fold][i]**2
        elif outscaler in ['minmax', 'robminmax']:
            Yhat[:, i] = scaler_resp[fold].inverse_transform(yhat, index=i)
            S2[:, i] = s2 * (scaler_resp[fold].max[i] - scaler_resp[fold].min[i])**2
        else:
            Yhat[:, i] = yhat.squeeze()
            S2[:, i] = s2.squeeze()

    if respfile is None:
        save_results(None, Yhat, S2, None, outputsuffix=outputsuffix)
        
        return (Yhat, S2)
    
    else:
        Y, maskvol = load_response_vars(respfile, maskfile)
        if models is not None and len(Y.shape) > 1:
            Y = Y[:, models]
            if meta_data:
                # are we using cross-validation?
                if type(mY) is list:
                    mY = mY[fold][models]
                else:
                    mY = mY[models]
                if type(sY) is list:
                    sY = sY[fold][models]
                else:
                    sY = sY[models]
        
        if len(Y.shape) == 1:
            Y = Y[:, np.newaxis]
            
        # warp the targets?   
        if alg == 'blr' and nm.blr.warp is not None:
            warp = True
            Yw = np.zeros_like(Y)            
            for i,m in enumerate(models):
                nm = norm_init(Xz)
                nm = nm.load(os.path.join(model_path, 'NM_' + str(fold) + '_' + 
                                          str(m) + inputsuffix + '.pkl'))

                warp_param = nm.blr.hyp[1:nm.blr.warp.get_n_params()+1] 
                Yw[:,i] = nm.blr.warp.f(Y[:,i], warp_param)
            Y = Yw;
        else:
            warp = False
        
        Z = (Y - Yhat) / np.sqrt(S2)
        
        print("Evaluating the model ...")
        if meta_data and not warp:
            
            results = evaluate(Y, Yhat, S2=S2, mY=mY, sY=sY)
        else:    
            results = evaluate(Y, Yhat, S2=S2, 
                           metrics = ['Rho', 'RMSE', 'SMSE', 'EXPV'])
        
        print("Evaluations Writing outputs ...")
        save_results(respfile, Yhat, S2, maskvol, Z=Z, 
                     outputsuffix=outputsuffix, results=results)
        
        return (Yhat, S2, Z)

    
def transfer(covfile, respfile, testcov=None, testresp=None, maskfile=None, 
             **kwargs):
    '''
    Transfer learning on the basis of a pre-estimated normative model by using 
    the posterior distribution over the parameters as an informed prior for 
    new data. currently only supported for HBR.
    
    Basic usage::

        transfer(covfile, respfile [extra_arguments])

    where the variables are defined below.

    :param covfile: transfer covariates used to predict the response variable
    :param respfile: transfer response variables for the normative model
    :param maskfile: mask used to apply to the data (nifti only)
    :param testcov: Test covariates
    :param testresp: Test responses
    :param model_path: Directory containing the normative model and metadata
    :param trbefile: Training batch effects file
    :param batch_size: batch size (for use with normative_parallel)
    :param job_id: batch id

    All outputs are written to disk in the same format as the input. These are:

    :outputs: * Yhat - predictive mean
              * S2 - predictive variance
              * Z - Z scores
    '''
    alg = kwargs.pop('alg').lower()
    
    if alg != 'hbr' and alg != 'blr':
        print('Model transfer function is only possible for HBR and BLR models.')
        return
    # testing should not be obligatory for HBR,
    # but should be for BLR (since it doesn't produce transfer models)
    elif (not 'model_path' in list(kwargs.keys()))  or \
        (not 'trbefile' in list(kwargs.keys())):
            print(f'{kwargs=}')
            print('InputError: Some general mandatory arguments are missing.')
            return
    # hbr has one additional mandatory arguments
    elif alg =='hbr':
        if (not 'output_path' in list(kwargs.keys())):
                print('InputError: Some mandatory arguments for hbr are missing.')
                return
        else: 
            output_path = kwargs.pop('output_path',None)
            if not os.path.isdir(output_path):
                os.mkdir(output_path) 

    # for hbr, testing is not mandatory, for blr's predict/transfer it is. This will be an architectural choice.
    #or (testresp==None)
    elif alg =='blr':
        if (testcov==None)   or \
        (not 'tsbefile' in list(kwargs.keys())):
                print('InputError: Some mandatory arguments for blr are missing.')
                return 
    # general arguments       
    log_path = kwargs.pop('log_path', None)
    model_path = kwargs.pop('model_path')
    outputsuffix = kwargs.pop('outputsuffix', 'transfer')
    outputsuffix = "_" + outputsuffix.replace("_", "")
    inputsuffix = kwargs.pop('inputsuffix', 'estimate')
    inputsuffix = "_" + inputsuffix.replace("_", "")
    tsbefile = kwargs.pop('tsbefile', None)
    trbefile = kwargs.pop('trbefile', None)
    job_id = kwargs.pop('job_id', None)
    batch_size = kwargs.pop('batch_size', None)
    fold = kwargs.pop('fold',0)
        
    # for PCNonline automated parallel jobs loop
    count_jobsdone = kwargs.pop('count_jobsdone','False')
    if type(count_jobsdone) is str:
        count_jobsdone = count_jobsdone=='True'
        
    if batch_size is not None:
        batch_size = int(batch_size)
        job_id = int(job_id) - 1
    
    if not os.path.isdir(model_path):
        print('Models directory does not exist!')
        return
    else:
        if os.path.exists(os.path.join(model_path, 'meta_data.md')):
            with open(os.path.join(model_path, 'meta_data.md'), 'rb') as file:
                meta_data = pickle.load(file)
            inscaler = meta_data['inscaler']
            outscaler = meta_data['outscaler']
            scaler_cov = meta_data['scaler_cov']
            scaler_resp = meta_data['scaler_resp']
            meta_data = True
        else:
            print("No meta-data file is found!")
            inscaler = 'None'
            outscaler = 'None'
            meta_data = False
       
    # load adaptation data
    print("Loading data ...")
    X = fileio.load(covfile)
    Y, maskvol = load_response_vars(respfile, maskfile)
    if len(Y.shape) == 1:
        Y = Y[:, np.newaxis]
    if len(X.shape) == 1:
        X = X[:, np.newaxis]
        
    if inscaler in ['standardize', 'minmax', 'robminmax']:
        X = scaler_cov[0].transform(X)
    
    feature_num = Y.shape[1]
    mY = np.mean(Y, axis=0)
    sY = np.std(Y, axis=0)  
    
    if outscaler in ['standardize', 'minmax', 'robminmax']:
        Y = scaler_resp[0].transform(Y)
    
    batch_effects_train = fileio.load(trbefile)
    
    # load test data
    if testcov is not None:
        # we have a separate test dataset
        Xte = fileio.load(testcov)
        if len(Xte.shape) == 1:
            Xte = Xte[:, np.newaxis]
        ts_sample_num = Xte.shape[0]
        if inscaler in ['standardize', 'minmax', 'robminmax']:
            Xte = scaler_cov[0].transform(Xte)
        
        if testresp is not None:
            Yte, testmask = load_response_vars(testresp, maskfile)
            if len(Yte.shape) == 1:
                Yte = Yte[:, np.newaxis]
        else:
            Yte = np.zeros([ts_sample_num, feature_num])
        
        if tsbefile is not None:
            batch_effects_test = fileio.load(tsbefile)
        else:
            batch_effects_test = np.zeros([Xte.shape[0],2])  
    else:
        ts_sample_num = 0     

    Yhat = np.zeros([ts_sample_num, feature_num])
    S2 = np.zeros([ts_sample_num, feature_num])
    Z = np.zeros([ts_sample_num, feature_num])
    
    # estimate the models for all subjects
    for i in range(feature_num):
              
        if alg == 'hbr':    
            print("Using HBR transform...")
            nm = norm_init(X)
            if batch_size is not None: # when using normative_parallel
                print("Transferring model ", job_id*batch_size+i)
                nm = nm.load(os.path.join(model_path, 'NM_0_' + 
                                          str(job_id*batch_size+i) + inputsuffix + 
                                          '.pkl'))
            else:
                print("Transferring model ", i+1, "of", feature_num)
                nm = nm.load(os.path.join(model_path, 'NM_0_' + str(i) + 
                                          inputsuffix + '.pkl'))
            
            nm = nm.estimate_on_new_sites(X, Y[:,i], batch_effects_train)
            if batch_size is not None: 
                nm.save(os.path.join(output_path, 'NM_0_' + 
                                 str(job_id*batch_size+i) + outputsuffix + '.pkl'))
            else:
                nm.save(os.path.join(output_path, 'NM_0_' + 
                                 str(i) + outputsuffix + '.pkl'))
            
            if testcov is not None:
                yhat, s2 = nm.predict_on_new_sites(Xte, batch_effects_test)
                
        # We basically use normative.predict script here.
        if alg == 'blr':
            print("Using BLR transform...")
            print("Transferring model ", i+1, "of", feature_num)      
            nm = norm_init(X)
            nm = nm.load(os.path.join(model_path, 'NM_' + str(fold) + '_' + 
                                      str(i) + inputsuffix + '.pkl'))
            
            # translate the syntax to what blr understands
            # first strip existing blr keyword arguments to avoid redundancy
            adapt_cov = kwargs.pop('adaptcovfile', None)
            adapt_res = kwargs.pop('adaptrespfile', None)
            adapt_vg = kwargs.pop('adaptvargroupfile', None)
            test_vg = kwargs.pop('testvargroupfile', None)
            if adapt_cov is not None or adapt_res is not None \
                or adapt_vg is not None or test_vg is not None:
                print("Warning: redundant batch effect parameterisation. Using HBR syntax")
            
            yhat, s2 = nm.predict(Xte, X, Y[:, i],
                                  adaptcov = X,
                                  adaptresp = Y[:, i],
                                  adaptvargroup = batch_effects_train,
                                  testvargroup = batch_effects_test,
                                  **kwargs)
        
        if testcov is not None:
            if outscaler == 'standardize': 
                Yhat[:, i] = scaler_resp[0].inverse_transform(yhat.squeeze(), index=i)
                S2[:, i] = s2.squeeze() * sY[i]**2
            elif outscaler in ['minmax', 'robminmax']:
                Yhat[:, i] = scaler_resp[0].inverse_transform(yhat, index=i)
                S2[:, i] = s2 * (scaler_resp[0].max[i] - scaler_resp[0].min[i])**2
            else:
                Yhat[:, i] = yhat.squeeze()
                S2[:, i] = s2.squeeze()
                
        
   
    if testresp is None:
        save_results(respfile, Yhat, S2, maskvol, outputsuffix=outputsuffix)
        return (Yhat, S2)
    else:
        # warp the targets?   
        if alg == 'blr' and nm.blr.warp is not None:
            warp = True
            Yw = np.zeros_like(Yte)            
            for i in range(feature_num):
                nm = norm_init(Xte)
                nm = nm.load(os.path.join(model_path, 'NM_' + str(fold) + '_' + 
                                          str(i) + inputsuffix + '.pkl'))

                warp_param = nm.blr.hyp[1:nm.blr.warp.get_n_params()+1] 
                Yw[:,i] = nm.blr.warp.f(Yte[:,i], warp_param)
            Yte = Yw;
        else:
            warp = False
            
        Z = (Yte - Yhat) / np.sqrt(S2)
    
        print("Evaluating the model ...")
        if meta_data and not warp:  
            results = evaluate(Yte, Yhat, S2=S2, mY=mY, sY=sY)
        else:    
            results = evaluate(Yte, Yhat, S2=S2, 
                           metrics = ['Rho', 'RMSE', 'SMSE', 'EXPV'])
                
        save_results(respfile, Yhat, S2, maskvol, Z=Z, results=results,
                     outputsuffix=outputsuffix)
        
        # Creates a file for every job succesfully completed (for tracking failed jobs).
        if count_jobsdone==True:
            done_path = os.path.join(log_path, str(job_id)+".jobsdone")
            Path(done_path).touch()
        
        return (Yhat, S2, Z)
    
    # Creates a file for every job succesfully completed (for tracking failed jobs).
    if count_jobsdone==True:
        done_path = os.path.join(log_path, str(job_id)+".jobsdone")
        Path(done_path).touch()


def extend(covfile, respfile, maskfile=None, **kwargs):
    
    '''
    This function extends an existing HBR model with data from new sites/scanners.
    
    Basic usage::

        extend(covfile, respfile [extra_arguments])

    where the variables are defined below.

    :param covfile: covariates for new data
    :param respfile: response variables for new data
    :param maskfile: mask used to apply to the data (nifti only)
    :param model_path: Directory containing the normative model and metadata
    :param trbefile: file address to batch effects file for new data
    :param batch_size: batch size (for use with normative_parallel)
    :param job_id: batch id
    :param output_path: the path for saving the  the extended model
    :param informative_prior: use initial model prior or learn from scratch (default is False).
    :param generation_factor: see below

    generation factor refers to the number of samples generated for each 
    combination of covariates and batch effects. Default is 10.


    All outputs are written to disk in the same format as the input.
    
    '''
    
    alg = kwargs.pop('alg')
    if alg != 'hbr':
        print('Model extention is only possible for HBR models.')
        return
    elif (not 'model_path' in list(kwargs.keys())) or \
        (not 'output_path' in list(kwargs.keys())) or \
        (not 'trbefile' in list(kwargs.keys())):
            print('InputError: Some mandatory arguments are missing.')
            return
    else:
        model_path = kwargs.pop('model_path')
        output_path = kwargs.pop('output_path')
        trbefile = kwargs.pop('trbefile')
    
    outputsuffix = kwargs.pop('outputsuffix', 'extend')
    outputsuffix = "_" + outputsuffix.replace("_", "")
    inputsuffix = kwargs.pop('inputsuffix', 'estimate')
    inputsuffix = "_" + inputsuffix.replace("_", "")
    informative_prior = kwargs.pop('informative_prior', 'False') == 'True'
    generation_factor = int(kwargs.pop('generation_factor', '10'))
    job_id = kwargs.pop('job_id', None)
    batch_size = kwargs.pop('batch_size', None)
    if batch_size is not None:
        batch_size = int(batch_size)
        job_id = int(job_id) - 1
     
    if not os.path.isdir(model_path):
        print('Models directory does not exist!')
        return
    else:
        if os.path.exists(os.path.join(model_path, 'meta_data.md')):
            with open(os.path.join(model_path, 'meta_data.md'), 'rb') as file:
                meta_data = pickle.load(file)
            if (meta_data['inscaler'] != 'None' or 
                meta_data['outscaler'] != 'None'):
                print('Models extention on scaled data is not possible!')
                return
    
    if not os.path.isdir(output_path):
        os.mkdir(output_path)
            
    # load data
    print("Loading data ...")
    X = fileio.load(covfile)    
    Y, maskvol = load_response_vars(respfile, maskfile)
    batch_effects_train = fileio.load(trbefile)
    
    if len(Y.shape) == 1:
        Y = Y[:, np.newaxis]
    if len(X.shape) == 1:
        X = X[:, np.newaxis]
    feature_num = Y.shape[1]
    
    # estimate the models for all subjects
    for i in range(feature_num):
              
        nm = norm_init(X)
        if batch_size is not None: # when using nirmative_parallel
            print("Extending model ", job_id*batch_size+i)
            nm = nm.load(os.path.join(model_path, 'NM_0_' + 
                                      str(job_id*batch_size+i) + inputsuffix + 
                                      '.pkl'))
        else:
            print("Extending model ", i+1, "of", feature_num)
            nm = nm.load(os.path.join(model_path, 'NM_0_' + str(i) + 
                                      inputsuffix +'.pkl'))
        
        nm = nm.extend(X, Y[:,i:i+1], batch_effects_train, 
                       samples=generation_factor, 
                       informative_prior=informative_prior)
        
        if batch_size is not None: 
            nm.save(os.path.join(output_path, 'NM_0_' + 
                             str(job_id*batch_size+i) + outputsuffix + '.pkl'))
            nm.save(os.path.join('Models', 'NM_0_' + 
                             str(i) + outputsuffix + '.pkl'))
        else:
            nm.save(os.path.join(output_path, 'NM_0_' + 
                             str(i) + outputsuffix + '.pkl'))
            
            

def tune(covfile, respfile, maskfile=None, **kwargs):
    
    '''
    This function tunes an existing HBR model with real data.
    
    Basic usage::

        tune(covfile, respfile [extra_arguments])

    where the variables are defined below.

    :param covfile: covariates for new data
    :param respfile: response variables for new data
    :param maskfile: mask used to apply to the data (nifti only)
    :param model_path: Directory containing the normative model and metadata
    :param trbefile: file address to batch effects file for new data
    :param batch_size: batch size (for use with normative_parallel)
    :param job_id: batch id
    :param output_path: the path for saving the  the extended model
    :param informative_prior: use initial model prior or learn from scracth (default is False).
    :param generation_factor: see below
    
    
    generation factor refers to the number of samples generated for each
    combination of covariates and batch effects. Default is 10.


    All outputs are written to disk in the same format as the input.
    
    '''
    
    alg = kwargs.pop('alg')
    if alg != 'hbr':
        print('Model extention is only possible for HBR models.')
        return
    elif (not 'model_path' in list(kwargs.keys())) or \
        (not 'output_path' in list(kwargs.keys())) or \
        (not 'trbefile' in list(kwargs.keys())):
            print('InputError: Some mandatory arguments are missing.')
            return
    else:
        model_path = kwargs.pop('model_path')
        output_path = kwargs.pop('output_path')
        trbefile = kwargs.pop('trbefile')
    
    outputsuffix = kwargs.pop('outputsuffix', 'tuned')
    outputsuffix = "_" + outputsuffix.replace("_", "")
    inputsuffix = kwargs.pop('inputsuffix', 'estimate')
    inputsuffix = "_" + inputsuffix.replace("_", "")
    informative_prior = kwargs.pop('informative_prior', 'False') == 'True'
    generation_factor = int(kwargs.pop('generation_factor', '10'))
    job_id = kwargs.pop('job_id', None)
    batch_size = kwargs.pop('batch_size', None)
    if batch_size is not None:
        batch_size = int(batch_size)
        job_id = int(job_id) - 1
     
    if not os.path.isdir(model_path):
        print('Models directory does not exist!')
        return
    else:
        if os.path.exists(os.path.join(model_path, 'meta_data.md')):
            with open(os.path.join(model_path, 'meta_data.md'), 'rb') as file:
                meta_data = pickle.load(file)
            if (meta_data['inscaler'] != 'None' or 
                meta_data['outscaler'] != 'None'):
                print('Models extention on scaled data is not possible!')
                return
    
    if not os.path.isdir(output_path):
        os.mkdir(output_path)
            
    # load data
    print("Loading data ...")
    X = fileio.load(covfile)    
    Y, maskvol = load_response_vars(respfile, maskfile)
    batch_effects_train = fileio.load(trbefile)
    
    if len(Y.shape) == 1:
        Y = Y[:, np.newaxis]
    if len(X.shape) == 1:
        X = X[:, np.newaxis]
    feature_num = Y.shape[1]
    
    # estimate the models for all subjects
    for i in range(feature_num):
              
        nm = norm_init(X)
        if batch_size is not None: # when using nirmative_parallel
            print("Tuning model ", job_id*batch_size+i)
            nm = nm.load(os.path.join(model_path, 'NM_0_' + 
                                      str(job_id*batch_size+i) + inputsuffix + 
                                      '.pkl'))
        else:
            print("Tuning model ", i+1, "of", feature_num)
            nm = nm.load(os.path.join(model_path, 'NM_0_' + str(i) + 
                                      inputsuffix +'.pkl'))
        
        nm = nm.tune(X, Y[:,i:i+1], batch_effects_train, 
                       samples=generation_factor, 
                       informative_prior=informative_prior)
        
        if batch_size is not None: 
            nm.save(os.path.join(output_path, 'NM_0_' + 
                             str(job_id*batch_size+i) + outputsuffix + '.pkl'))
            nm.save(os.path.join('Models', 'NM_0_' + 
                             str(i) + outputsuffix + '.pkl'))
        else:
            nm.save(os.path.join(output_path, 'NM_0_' + 
                             str(i) + outputsuffix + '.pkl'))


def merge(covfile=None, respfile=None, **kwargs):
    
    '''
    This function extends an existing HBR model with data from new sites/scanners.
    
    Basic usage::

        merge(model_path1, model_path2 [extra_arguments])

    where the variables are defined below.

    :param covfile: Not required. Always set to None.
    :param respfile: Not required. Always set to None.
    :param model_path1: Directory containing the model and metadata (1st model)
    :param model_path2: Directory containing the  model and metadata (2nd model)
    :param batch_size: batch size (for use with normative_parallel)
    :param job_id: batch id
    :param output_path: the path for saving the  the extended model
    :param generation_factor: see below

    The generation factor refers tothe number of samples generated for each 
    combination of covariates and batch effects. Default is 10.


    All outputs are written to disk in the same format as the input.
    
    '''
    
    alg = kwargs.pop('alg')
    if alg != 'hbr':
        print('Merging models is only possible for HBR models.')
        return
    elif (not 'model_path1' in list(kwargs.keys())) or \
        (not 'model_path2' in list(kwargs.keys())) or \
        (not 'output_path' in list(kwargs.keys())):
            print('InputError: Some mandatory arguments are missing.')
            return
    else:
        model_path1 = kwargs.pop('model_path1')
        model_path2 = kwargs.pop('model_path2')
        output_path = kwargs.pop('output_path')
    
    outputsuffix = kwargs.pop('outputsuffix', 'merge')
    outputsuffix = "_" + outputsuffix.replace("_", "")
    inputsuffix = kwargs.pop('inputsuffix', 'estimate')
    inputsuffix = "_" + inputsuffix.replace("_", "")
    generation_factor = int(kwargs.pop('generation_factor', '10'))
    job_id = kwargs.pop('job_id', None)
    batch_size = kwargs.pop('batch_size', None)
    if batch_size is not None:
        batch_size = int(batch_size)
        job_id = int(job_id) - 1
     
    if (not os.path.isdir(model_path1)) or (not os.path.isdir(model_path2)):
        print('Models directory does not exist!')
        return
    else:
        if batch_size is None:
            with open(os.path.join(model_path1, 'meta_data.md'), 'rb') as file:
                meta_data1 = pickle.load(file)
            with open(os.path.join(model_path2, 'meta_data.md'), 'rb') as file:
                meta_data2 = pickle.load(file)
            if meta_data1['valid_voxels'].shape[0] != meta_data2['valid_voxels'].shape[0]:
                print('Two models are trained on different features!')
                return
            else:
                feature_num = meta_data1['valid_voxels'].shape[0]
        else:
            feature_num = batch_size
            
            
    if not os.path.isdir(output_path):
        os.mkdir(output_path)
    
    # mergeing the models
    for i in range(feature_num):
              
        nm1 = norm_init(np.random.rand(100,10))
        nm2 = norm_init(np.random.rand(100,10))
        if batch_size is not None: # when using nirmative_parallel
            print("Merging model ", job_id*batch_size+i)
            nm1 = nm1.load(os.path.join(model_path1, 'NM_0_' + 
                                      str(job_id*batch_size+i) + inputsuffix + 
                                      '.pkl'))
            nm2 = nm2.load(os.path.join(model_path2, 'NM_0_' + 
                                      str(job_id*batch_size+i) + inputsuffix + 
                                      '.pkl'))
        else:
            print("Merging model ", i+1, "of", feature_num)
            nm1 = nm1.load(os.path.join(model_path1, 'NM_0_' + str(i) + 
                                      inputsuffix +'.pkl'))
            nm2 = nm1.load(os.path.join(model_path2, 'NM_0_' + str(i) + 
                                      inputsuffix +'.pkl'))
        
        nm_merged = nm1.merge(nm2, samples=generation_factor)
        
        if batch_size is not None: 
            nm_merged.save(os.path.join(output_path, 'NM_0_' + 
                             str(job_id*batch_size+i) + outputsuffix + '.pkl'))
            nm_merged.save(os.path.join('Models', 'NM_0_' + 
                             str(i) + outputsuffix + '.pkl'))
        else:
            nm_merged.save(os.path.join(output_path, 'NM_0_' + 
                             str(i) + outputsuffix + '.pkl'))


def main(*args):
    """ Parse arguments and estimate model
    """

    np.seterr(invalid='ignore')

    rfile, mfile, cfile, cv, tcfile, trfile, func, alg, cfg, kw = get_args(args)
    
    # collect required arguments
    pos_args = ['cfile', 'rfile']
    
    # collect basic keyword arguments controlling model estimation
    kw_args = ['maskfile=mfile',
               'cvfolds=cv',
               'testcov=tcfile',
               'testresp=trfile',
               'alg=alg',
               'configparam=cfg']
    
    # add additional keyword arguments
    for k in kw:
        kw_args.append(k + '=' + "'" + kw[k] + "'")
    all_args = ', '.join(pos_args + kw_args)

    # Executing the target function
    exec(func + '(' + all_args + ')')

# For running from the command line:
if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 24 15:06:06 2019

@author: seykia
"""

import torch
from torch import nn
from torch.nn import functional as F

##################################### NP Model ################################

class NP(nn.Module):
    def __init__(self, encoder, decoder, args):
        super(NP, self).__init__()
        self.r_dim = encoder.r_dim
        self.z_dim = encoder.z_dim
        self.dp_level = encoder.dp_level
        self.encoder = encoder
        self.decoder = decoder
        self.r_to_z_mean_dp = nn.Dropout(p = self.dp_level)
        self.r_to_z_mean = nn.Linear(self.r_dim, self.z_dim)
        self.r_to_z_logvar_dp = nn.Dropout(p = self.dp_level)
        self.r_to_z_logvar = nn.Linear(self.r_dim, self.z_dim)
        self.device = args.device
        self.type = args.type
        
    def xy_to_z_params(self, x, y):
        r = self.encoder.forward(x, y)
        mu = self.r_to_z_mean(self.r_to_z_mean_dp(r))
        logvar = self.r_to_z_logvar(self.r_to_z_logvar_dp(r))
        return mu, logvar
      
    def reparameterise(self, z):
        mu, logvar = z
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        z_sample = eps.mul(std).add_(mu)
        return z_sample

    def forward(self, x_context, y_context, x_all=None, y_all=None, n = 10):
        y_sigma = None
        z_context = self.xy_to_z_params(x_context, y_context)
        if self.training:
            z_all = self.xy_to_z_params(x_all, y_all)
            z_sample = self.reparameterise(z_all)
            y_hat = self.decoder.forward(z_sample, x_all)
        else:  
            z_all = z_context
            if self.type == 'ST':
                temp = torch.zeros([n,y_context.shape[0], y_context.shape[2]], device = 'cpu')
            elif self.type == 'MT':
                temp = torch.zeros([n,y_context.shape[0],1,y_context.shape[2],y_context.shape[3],
                                y_context.shape[4]], device = 'cpu')                                
            for i in range(n):
                z_sample = self.reparameterise(z_all)
                temp[i,:] = self.decoder.forward(z_sample, x_context)
            y_hat = torch.mean(temp, dim=0).to(self.device)
            if n > 1:
                y_sigma = torch.std(temp, dim=0).to(self.device)
        return y_hat, z_all, z_context, y_sigma
    
###############################################################################
        
def apply_dropout_test(m):
    if type(m) == nn.Dropout:
        m.train()

def kl_div_gaussians(mu_q, logvar_q, mu_p, logvar_p):
    var_p = torch.exp(logvar_p)
    kl_div = (torch.exp(logvar_q) + (mu_q - mu_p) ** 2) / (var_p) \
             - 1.0 \
             + logvar_p - logvar_q
    kl_div = 0.5 * kl_div.sum()
    return kl_div

def np_loss(y_hat, y, z_all, z_context):
    BCE = F.binary_cross_entropy(torch.squeeze(y_hat), torch.mean(y,dim=1), reduction="sum")
    KLD = kl_div_gaussians(z_all[0], z_all[1], z_context[0], z_context[1])
    return BCE + KLD



from __future__ import print_function
from __future__ import division

import numpy as np
import torch

class GPRRFA:
    """Random Feature Approximation for Gaussian Process Regression

    Estimation and prediction of Bayesian linear regression models

    Basic usage::

        R = GPRRFA()
        hyp = R.estimate(hyp0, X, y)
        ys,s2 = R.predict(hyp, X, y, Xs)

    where the variables are

    :param hyp: vector of hyperparmaters.
    :param X: N x D data array
    :param y: 1D Array of targets (length N)
    :param Xs: Nte x D array of test cases
    :param hyp0: starting estimates for hyperparameter optimisation

    :returns: * ys - predictive mean
              * s2 - predictive variance

    The hyperparameters are::

        hyp = [ log(sn), log(ell), log(sf) ]  # hyp is a numpy array

    where sn^2 is the noise variance, ell are lengthscale parameters and 
    sf^2 is the signal variance. This provides an approximation to the
    covariance function::
        
        k(x,z) = x'*z + sn2*exp(0.5*(x-z)'*Lambda*(x-z))
    
    where Lambda = diag((ell_1^2, ... ell_D^2))

    Written by A. Marquand
    """

    def __init__(self, hyp=None, X=None, y=None, n_feat=None,
                 n_iter=100, tol=1e-3, verbose=False):

        self.hyp = np.nan
        self.nlZ = np.nan
        self.tol = tol          # not used at present
        self.Nf = n_feat
        self.n_iter = n_iter
        self.verbose = verbose
        self._n_restarts = 5

        if (hyp is not None) and (X is not None) and (y is not None):
            self.post(hyp, X, y)

    def _numpy2torch(self, X, y=None, hyp=None):

        if type(X) is torch.Tensor:
           pass
        elif type(X) is np.ndarray:
           X = torch.from_numpy(X)
        else:
           raise(ValueError, 'Unknown data type (X)')
        X = X.double()
        
        if y is not None:
            if type(y) is torch.Tensor:
                pass
            elif type(y) is np.ndarray:
                y = torch.from_numpy(y)
            else:
                raise(ValueError, 'Unknown data type (y)')
            
            if len(y.shape) == 1:
                y.resize_(y.shape[0],1)
            y = y.double()
        
        if hyp is not None:
            if type(hyp) is torch.Tensor:
                pass
            else:
                hyp = torch.tensor(hyp, requires_grad=True)
        
        return X, y, hyp
    
    def get_n_params(self, X):
        
        return X.shape[1] + 2
        
    def post(self, hyp, X, y):
        """ Generic function to compute posterior distribution.

            This function will save the posterior mean and precision matrix as
            self.m and self.A and will also update internal parameters (e.g.
            N, D and the prior covariance (Sigma) and precision (iSigma).
        """
     
        # make sure all variables are the right type
        X, y, hyp = self._numpy2torch(X, y, hyp)
        
        self.N, self.Dx = X.shape
        
        # ensure the number of features is specified (use 75% as a default)
        if self.Nf is None:
            self.Nf = int(0.75 * self.N)
        
        self.Omega = torch.zeros((self.Dx, self.Nf), dtype=torch.double)
        for f in range(self.Nf):
            self.Omega[:,f] = torch.exp(hyp[1:-1]) * \
            torch.randn((self.Dx, 1), dtype=torch.double).squeeze()

        XO = torch.mm(X, self.Omega) 
        self.Phi = torch.exp(hyp[-1])/np.sqrt(self.Nf) *  \
                   torch.cat((torch.cos(XO), torch.sin(XO)), 1)
        
        # concatenate linear weights 
        self.Phi = torch.cat((self.Phi, X), 1)
        self.D = self.Phi.shape[1]

        if self.verbose:
            print("estimating posterior ... | hyp=", hyp)
        
        self.A = torch.mm(torch.t(self.Phi), self.Phi) / torch.exp(2*hyp[0]) + \
                 torch.eye(self.D, dtype=torch.double)
        self.m = torch.mm(torch.solve(torch.t(self.Phi), self.A)[0], y) / \
                 torch.exp(2*hyp[0])

        # save hyperparameters
        self.hyp = hyp
        
        # update optimizer iteration count
        if hasattr(self,'_iterations'):
            self._iterations += 1

    def loglik(self, hyp, X, y):
        """ Function to compute compute log (marginal) likelihood """
        X, y, hyp = self._numpy2torch(X, y, hyp)

        # always recompute the posterior
        self.post(hyp, X, y)

        #logdetA = 2*torch.sum(torch.log(torch.diag(torch.cholesky(self.A))))
        try:
            # compute the log determinants in a numerically stable way
            logdetA = 2*torch.sum(torch.log(torch.diag(torch.cholesky(self.A))))
        except Exception as e:
            print("Warning: Estimation of posterior distribution failed")
            print(e)
            #nlZ = torch.tensor(1/np.finfo(float).eps)
            nlZ = torch.tensor(np.nan)
            self._optim_failed = True
            return nlZ
        
        # compute negative marginal log likelihood
        nlZ = -0.5 * (self.N*torch.log(1/torch.exp(2*hyp[0])) - 
                      self.N*np.log(2*np.pi) -
                      torch.mm(torch.t(y - torch.mm(self.Phi,self.m)),
                               (y - torch.mm(self.Phi,self.m))) / 
                      torch.exp(2*hyp[0]) -
                      torch.mm(torch.t(self.m), self.m) - logdetA)

        if self.verbose:
            print("nlZ= ", nlZ, " | hyp=", hyp)

        # save marginal likelihood
        self.nlZ = nlZ
        return nlZ

    def dloglik(self, hyp, X, y):
        """ Function to compute derivatives """

        print("derivatives not available")

        return

    def estimate(self, hyp0, X, y, optimizer='lbfgs'):
        """ Function to estimate the model """
        
        if type(hyp0) is torch.Tensor:
            hyp = hyp0
            hyp0.requires_grad_()
        else:
            hyp = torch.tensor(hyp0, requires_grad=True) 
        # save the starting values
        self.hyp0 = hyp
        
        if optimizer.lower() == 'lbfgs':
            opt = torch.optim.LBFGS([hyp])
        else:
            raise(ValueError, "Optimizer " + " not implemented")
        self._iterations = 0
        
        def closure():
            opt.zero_grad()
            nlZ = self.loglik(hyp, X, y)
            if not torch.isnan(nlZ):
                nlZ.backward()
            return nlZ
        
        for r in range(self._n_restarts):
            self._optim_failed = False
            
            nlZ = opt.step(closure)
            
            if self._optim_failed:
                print("optimization failed. retrying (", r+1, "of", 
                      self._n_restarts,")")
                hyp = torch.randn_like(hyp, requires_grad=True)
                self.hyp0 = hyp
            else:
                print("Optimzation complete after", self._iterations, 
                      "evaluations. Function value =", 
                      nlZ.detach().numpy().squeeze())
                break

        return self.hyp.detach().numpy()

    def predict(self, hyp, X, y, Xs):
        """ Function to make predictions from the model """

        X, y, hyp = self._numpy2torch(X, y, hyp)
        Xs, *_ = self._numpy2torch(Xs)

        if (hyp != self.hyp).all() or not(hasattr(self, 'A')):
            self.post(hyp, X, y)
        
        # generate prediction tensors
        XsO = torch.mm(Xs, self.Omega) 
        Phis = torch.exp(hyp[-1])/np.sqrt(self.Nf) * \
               torch.cat((torch.cos(XsO), torch.sin(XsO)), 1)
        # add linear component
        Phis = torch.cat((Phis, Xs), 1)
        
        ys = torch.mm(Phis, self.m)

        # compute diag(Phis*(Phis'\A)) avoiding computing off-diagonal entries
        s2 = torch.exp(2*hyp[0]) + \
                torch.sum(Phis * torch.t(torch.solve(torch.t(Phis), self.A)[0]), 1)

        # return output as numpy arrays
        return ys.detach().numpy().squeeze(), s2.detach().numpy().squeeze()



import theano.tensor
from pymc3.distributions import Continuous, draw_values, generate_samples
import theano.tensor as tt
import numpy as np
from pymc3.distributions.dist_math import bound
import scipy.special as spp
from theano import as_op
from theano.gof.fg import NullType
from theano.gof.op import Op
from theano.gof.graph import Apply
from theano.gradient import grad_not_implemented

"""
@author: Stijn de Boer (AuguB)

See: Jones et al. (2009), Sinh-Arcsinh distributions.
"""


class K(Op):
    """
    Modified Bessel function of the second kind, theano implementation
    """
    def make_node(self, p, x):
        p = theano.tensor.as_tensor_variable(p, 'floatX')
        x = theano.tensor.as_tensor_variable(x, 'floatX')
        return Apply(self, [p,x], [p.type()])

    def perform(self, node, inputs, output_storage, params=None):
        # Doing this on the unique values avoids doing A LOT OF double work, apparently scipy doesn't do this by itself
        unique_inputs, inverse_indices = np.unique(inputs[0], return_inverse=True)
        unique_outputs = spp.kv(unique_inputs, inputs[1])
        outputs = unique_outputs[inverse_indices].reshape(inputs[0].shape)
        output_storage[0][0] = outputs

    def grad(self, inputs, output_grads):
        # Approximation of the derivative. This should suffice for using NUTS
        dp = 1e-10
        p = inputs[0]
        x = inputs[1]
        grad = (self(p+dp,x) - self(p, x))/dp
        return [output_grads[0]*grad, grad_not_implemented(0,1,2,3)]

class SHASH(Continuous):
    """
    SHASH described by Jones et al., based on a standard normal
    All SHASH subclasses inherit from this
    """
    def __init__(self, epsilon, delta, **kwargs):
        super().__init__(**kwargs)
        self.epsilon = tt.as_tensor_variable(epsilon)
        self.delta = tt.as_tensor_variable(delta)
        self.K = K()


    def random(self, point=None, size=None):
        epsilon, delta = draw_values([self.epsilon, self.delta],
                                     point=point, size=size)

        def _random(epsilon, delta, size=None):
            samples_transformed = np.sinh((np.arcsinh(np.random.randn(*size)) + epsilon) / delta)
            return samples_transformed

        return generate_samples(_random, epsilon=epsilon, delta=delta, dist_shape=self.shape, size=size)

    def logp(self, value):
        epsilon = self.epsilon
        delta = self.delta + tt.np.finfo(np.float32).eps

        this_S = self.S(value)
        this_S_sqr = tt.sqr(this_S)
        this_C_sqr = 1+this_S_sqr
        frac1 = -tt.log(tt.constant(2 * tt.np.pi))/2
        frac2 = tt.log(delta) + tt.log(this_C_sqr)/2 - tt.log(1 + tt.sqr(value)) / 2
        exp = -this_S_sqr / 2

        return bound(frac1 + frac2 + exp, delta > 0)

    def S(self, x):
        """

        :param epsilon:
        :param delta:
        :param x:
        :return: The sinharcsinh transformation of x
        """
        return tt.sinh(tt.arcsinh(x) * self.delta - self.epsilon)

    def S_inv(self, x):
        return tt.sinh((tt.arcsinh(x) + self.epsilon) / self.delta)

    def C(self, x):
        """
        :param epsilon:
        :param delta:
        :param x:
        :return: the cosharcsinh transformation of x
        Be aware that this is sqrt(1+S(x)^2), so you may save some compute if you can re-use the result from S.
        """
        return tt.cosh(tt.arcsinh(x) * self.delta - self.epsilon)

    def P(self, q):
        """
        The P function as given in Jones et al.
        :param q:
        :return:
        """
        frac = np.exp(1 / 4) / np.power(8 * np.pi, 1 / 2)
        K1 = self.K((q+1)/2,1/4)
        K2 = self.K((q-1)/2,1/4)
        a = (K1 + K2) * frac
        return a

    def m(self, r):
        """
        :param epsilon:
        :param delta:
        :param r:
        :return:  The r'th uncentered moment of the SHASH distribution parameterized by epsilon and delta. Given by Jones et al.
        """
        frac1 = tt.as_tensor_variable(1 / np.power(2, r))
        acc = tt.as_tensor_variable(0)
        for i in range(r + 1):
            combs = spp.comb(r, i)
            flip = np.power(-1, i)
            ex = np.exp((r - 2 * i) * self.epsilon / self.delta)
            # This is the reason we can not sample delta/kurtosis using NUTS; the gradient of P is unknown to pymc3
            # TODO write a class that inherits theano.Op and do the gradient in there :)
            p = self.P((r - 2 * i) / self.delta)
            acc += combs * flip * ex * p
        return frac1 * acc

class SHASHo(SHASH):
    """
    This is the shash where the location and scale parameters have simply been applied as an linear transformation
    directly on the original shash.
    """

    def __init__(self, mu, sigma, epsilon, delta, **kwargs):
        super().__init__(epsilon, delta, **kwargs)
        self.mu = tt.as_tensor_variable(mu)
        self.sigma = tt.as_tensor_variable(sigma)

    def random(self, point=None, size=None):
        mu, sigma, epsilon, delta = draw_values([self.mu, self.sigma, self.epsilon, self.delta],
                                                point=point, size=size)

        def _random(mu, sigma, epsilon, delta, size=None):
            samples_transformed = np.sinh((np.arcsinh(np.random.randn(*size)) + epsilon) / delta) * sigma + mu
            return samples_transformed

        return generate_samples(_random, mu=mu, sigma=sigma, epsilon=epsilon, delta=delta,
                                dist_shape=self.shape,
                                size=size)

    def logp(self, value):
        mu = self.mu
        sigma = self.sigma + tt.np.finfo(np.float32).eps
        epsilon = self.epsilon
        delta = self.delta + tt.np.finfo(np.float32).eps

        value_transformed = (value - mu) / sigma

        this_S = self.S( value_transformed)
        this_S_sqr = tt.sqr(this_S)
        this_C_sqr = 1+this_S_sqr
        frac1 = -tt.log(tt.constant(2 * tt.np.pi))/2
        frac2 = tt.log(delta) + tt.log(this_C_sqr)/2 - tt.log(
            1 + tt.sqr(value_transformed)) / 2
        exp = -this_S_sqr / 2
        change_of_variable = -tt.log(sigma)

        return bound(frac1 + frac2 + exp + change_of_variable, sigma > 0, delta > 0)


class SHASHo2(SHASH):
    """
    This is the shash where we apply the reparameterization provided in section 4.3 in Jones et al.
    """

    def __init__(self, mu, sigma, epsilon, delta, **kwargs):
        super().__init__(epsilon, delta, **kwargs)
        self.mu = tt.as_tensor_variable(mu)
        self.sigma = tt.as_tensor_variable(sigma)

    def random(self, point=None, size=None):
        mu, sigma, epsilon, delta = draw_values(
            [self.mu, self.sigma, self.epsilon, self.delta],
            point=point, size=size)
        sigma_d = sigma / delta

        def _random(mu, sigma, epsilon, delta, size=None):
            samples_transformed = np.sinh(
                (np.arcsinh(np.random.randn(*size)) + epsilon) / delta) * sigma_d + mu
            return samples_transformed

        return generate_samples(_random, mu=mu, sigma=sigma_d, epsilon=epsilon, delta=delta,
                                dist_shape=self.shape,
                                size=size)

    def logp(self, value):
        mu = self.mu
        sigma = self.sigma + tt.np.finfo(np.float32).eps
        epsilon = self.epsilon
        delta = self.delta + tt.np.finfo(np.float32).eps
        sigma_d = sigma / delta


        # Here a double change of variables is applied
        value_transformed = ((value - mu) / sigma_d)

        this_S = self.S(value_transformed)
        this_S_sqr = tt.sqr(this_S)
        this_C = tt.sqrt(1+this_S_sqr)
        frac1 = -tt.log(tt.sqrt(tt.constant(2 * tt.np.pi)))
        frac2 = tt.log(delta) + tt.log(this_C) - tt.log(
            1 + tt.sqr(value_transformed)) / 2
        exp = -this_S_sqr / 2
        change_of_variable = -tt.log(sigma_d)

        # the change of variables is accounted for in the density by division and multiplication (adding and subtracting for logp)
        return bound(frac1 + frac2 + exp + change_of_variable, delta > 0, sigma > 0)

class SHASHb(SHASH):
    """
    This is the shash where the location and scale parameters been applied as an linear transformation on the shash
    distribution which was corrected for mean and variance.
    """

    def __init__(self, mu, sigma, epsilon, delta, **kwargs):
        super().__init__(epsilon, delta, **kwargs)
        self.mu = tt.as_tensor_variable(mu)
        self.sigma = tt.as_tensor_variable(sigma)

    def random(self, point=None, size=None):
        mu, sigma, epsilon, delta = draw_values(
            [self.mu, self.sigma, self.epsilon, self.delta],
            point=point, size=size)
        mean = (tt.sinh(epsilon/delta)*self.P(1/delta)).eval()
        var =  ((tt.cosh(2*epsilon/delta)*self.P(2/delta)-1)/2).eval() - mean**2

        def _random(mean, var, mu, sigma, epsilon, delta, size=None):
            samples_transformed = ((np.sinh(
                (np.arcsinh(np.random.randn(*size)) + epsilon) / delta) - mean) / np.sqrt(var)) * sigma + mu
            return samples_transformed

        return generate_samples(_random, mean=mean, var=var, mu=mu, sigma=sigma, epsilon=epsilon, delta=delta,
                                dist_shape=self.shape,
                                size=size)

    def logp(self, value):
        mu = self.mu
        sigma = self.sigma + tt.np.finfo(np.float32).eps
        epsilon = self.epsilon
        delta = self.delta + tt.np.finfo(np.float32).eps
        mean = tt.sinh(epsilon/delta)*self.P(1/delta)
        var = (tt.cosh(2*epsilon/delta)*self.P(2/delta)-1)/2 - tt.sqr(mean)

        # Here a double change of variables is applied
        value_transformed = ((value - mu) / sigma) * tt.sqrt(var) + mean

        this_S = self.S(value_transformed)
        this_S_sqr = tt.sqr(this_S)
        this_C_sqr = 1+this_S_sqr
        frac1 = -tt.log(tt.constant(2 * tt.np.pi))/2
        frac2 = tt.log(delta) + tt.log(this_C_sqr)/2 - tt.log(1 + tt.sqr(value_transformed)) / 2
        exp = -this_S_sqr / 2
        change_of_variable = tt.log(var)/2 - tt.log(sigma)

        # the change of variables is accounted for in the density by division and multiplication (addition and subtraction in the log domain)
        return bound(frac1 + frac2 + exp + change_of_variable, delta > 0, sigma > 0, var > 0)



from setuptools import setup, find_packages

setup(name='pcntoolkit',
      version='0.27',
      description='Predictive Clinical Neuroscience toolkit',
      url='http://github.com/amarquand/PCNtoolkit',
      author='Andre Marquand',
      author_email='andre.marquand@donders.ru.nl',
      license='GNU GPLv3',
      packages=find_packages(),
      install_requires=[
          'argparse',
          'nibabel>=2.5.1',
          'six',
          'scikit-learn', 
          'bspline',
          'matplotlib',
          'numpy>=1.19.5,<1.23',
          'scipy>=1.3.2',
          'pandas>=0.25.3',
          'torch>=1.1.0', 
          'sphinx-tabs',
          'pymc3>=3.8,<=3.9.3',
          'theano==1.0.5',
          'arviz==0.11.0'
      ],
      #python_requires='<3.10',
      zip_safe=False)



import os
import sys
from six import with_metaclass
from abc import ABCMeta, abstractmethod
import pickle

try:  # run as a package if installed
    from pcntoolkit import configs
except ImportError:
    pass

    path = os.path.abspath(os.path.dirname(__file__))
    if path not in sys.path:
        sys.path.append(path)
    del path
    import configs


class NormBase(with_metaclass(ABCMeta)):
    """ Base class for normative model back-end.

        All normative modelling approaches must define the following methods::

            NormativeModel.estimate()
            NormativeModel.predict()
    """

    def __init__(self, x=None):
        pass

    @abstractmethod
    def estimate(self, X, y):
        """ Estimate the normative model """

    @abstractmethod
    def predict(self, Xs, X, y):
        """ Make predictions for new data """

    @property
    @abstractmethod
    def n_params(self):
        """ Report the number of parameters required by the model """
    
    def save(self, save_path):
        try:
            with open(save_path, 'wb') as handle:
                pickle.dump(self, handle, protocol=configs.PICKLE_PROTOCOL)
            return True
        except Exception as err:
            print('Error:', err)
            raise
    
    def load(self, load_path):
        try:
            with open(load_path, 'rb') as handle:
                nm = pickle.load(handle)
            return nm
        except Exception as err:
            print('Error:', err)
            raise
